import pytest, warnings, numpy as np
from sequentia.classifiers import _Topology, _LeftRightTopology, _ErgodicTopology, _LinearTopology
from ....support import assert_equal, assert_all_equal, assert_distribution

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# ========= #
# _Topology #
# ========= #

# --------------------------- #
# _Topology.uniform_initial() #
# --------------------------- #

def test_uniform_initial_min():
    """Generate a uniform initial state distribution with the minimum number of states"""
    topology = _Topology(n_states=1, random_state=rng)
    initial = topology.uniform_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        1.
    ]))

def test_uniform_initial_small():
    """Generate a uniform initial state distribution with a few states"""
    topology = _Topology(n_states=2, random_state=rng)
    initial = topology.uniform_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.5, 0.5
    ]))

def test_uniform_initial_many():
    """Generate a uniform initial state distribution with many states"""
    topology = _Topology(n_states=5, random_state=rng)
    initial = topology.uniform_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

# -------------------------- #
# _Topology.random_initial() #
# -------------------------- #

def test_random_initial_min():
    """Generate a random initial state distribution with minimal states"""
    topology = _Topology(n_states=1, random_state=rng)
    initial = topology.random_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        1.
    ]))

def test_random_initial_small():
    """Generate a random initial state distribution with few states"""
    topology = _Topology(n_states=2, random_state=rng)
    initial = topology.random_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.57633871, 0.42366129
    ]))

def test_random_initial_many():
    """Generate a random initial state distribution with many states"""
    topology = _Topology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    assert_distribution(initial)
    assert_equal(initial, np.array([
        0.15210286, 0.10647349, 0.20059295, 0.11120171, 0.42962898
    ]))

# ================== #
# _LeftRightTopology #
# ================== #

# ---------------------------------------- #
# _LeftRightTopology.uniform_transitions() #
# ---------------------------------------- #

def test_left_right_uniform_transitions_min():
    """Generate a uniform left-right transition matrix with minimal states"""
    topology = _LeftRightTopology(n_states=1, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_left_right_uniform_transitions_small():
    """Generate a uniform left-right transition matrix with few states"""
    topology = _LeftRightTopology(n_states=2, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5],
        [0. , 1. ]
    ]))

def test_left_right_uniform_transitions_many():
    """Generate a uniform left-right transition matrix with many states"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.2, 0.2 , 0.2       , 0.2       , 0.2       ],
        [0. , 0.25, 0.25      , 0.25      , 0.25      ],
        [0. , 0.  , 0.33333333, 0.33333333, 0.33333333],
        [0. , 0.  , 0.        , 0.5       , 0.5       ] ,
        [0. , 0.  , 0.        , 0.        , 1.        ]
    ]))

# --------------------------------------- #
# _LeftRightTopology.random_transitions() #
# --------------------------------------- #

def test_left_right_random_transitions_min():
    """Generate a random left-right transition matrix with minimal states"""
    topology = _LeftRightTopology(n_states=1, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_left_right_random_transitions_small():
    """Generate a random left-right transition matrix with few states"""
    topology = _LeftRightTopology(n_states=2, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.23561633, 0.76438367],
        [0.        , 1.        ]
    ]))

def test_left_right_random_transitions_many():
    """Generate a random left-right transition matrix with many states"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.56841967, 0.01612013, 0.01994328, 0.0044685 , 0.39104841],
        [0.        , 0.25134034, 0.43904868, 0.20609306, 0.10351793],
        [0.        , 0.        , 0.27462001, 0.12291279, 0.60246721],
        [0.        , 0.        , 0.        , 0.61951739, 0.38048261],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

# ----------------------------------------- #
# _LeftRightTopology.validate_transitions() #
# ----------------------------------------- #

def test_left_right_validate_transitions_invalid():
    """Validate an invalid left-right transition matrix"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = _ErgodicTopology(n_states=5, random_state=rng).random_transitions()
    with pytest.raises(ValueError) as e:
        topology.validate_transitions(transitions)
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_left_right_validate_transitions_valid():
    """Validate a valid left-right transition matrix"""
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    topology.validate_transitions(transitions)

# -------------------------------------- #
# _ErgodicTopology.uniform_transitions() #
# -------------------------------------- #

def test_ergodic_uniform_transitions_min():
    """Generate a uniform ergodic transition matrix with minimal states"""
    topology = _ErgodicTopology(n_states=1, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_ergodic_uniform_transitions_small():
    """Generate a uniform ergodic transition matrix with few states"""
    topology = _ErgodicTopology(n_states=2, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5],
        [0.5, 0.5]
    ]))

def test_ergodic_uniform_transitions_many():
    """Generate a uniform ergodic transition matrix with many states"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))

# ------------------------------------- #
# _ErgodicTopology.random_transitions() #
# ------------------------------------- #

def test_ergodic_random_transitions_min():
    """Generate a random ergodic transition matrix with minimal states"""
    topology = _ErgodicTopology(n_states=1, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_ergodic_random_transitions_small():
    """Generate a random ergodic transition matrix with few states"""
    topology = _ErgodicTopology(n_states=2, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.87353002, 0.12646998],
        [0.88622334, 0.11377666]
    ]))

def test_ergodic_random_transitions_many():
    """Generate a random ergodic transition matrix with many states"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.46537016, 0.12619365, 0.07474032, 0.32619324, 0.00750262],
        [0.38836848, 0.00103519, 0.24911885, 0.06922191, 0.29225557],
        [0.5312161 , 0.04639154, 0.13922816, 0.14542372, 0.13774047],
        [0.0361995 , 0.43772711, 0.08498809, 0.26867251, 0.17241279],
        [0.06373359, 0.30347054, 0.09117514, 0.38445582, 0.1571649 ]
    ]))

# --------------------------------------- #
# _ErgodicTopology.validate_transitions() #
# --------------------------------------- #

def test_ergodic_validate_transitions_invalid():
    """Validate an invalid ergodic transition matrix"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = _LeftRightTopology(n_states=5, random_state=rng).random_transitions()
    with pytest.warns(UserWarning):
        topology.validate_transitions(transitions)

def test_ergodic_validate_transitions_valid():
    """Validate a valid ergodic transition matrix"""
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    topology.validate_transitions(transitions)

# ======================== #
# _LinearTopology #
# ======================== #

# ---------------------------------------------- #
# _LinearTopology.uniform_transitions() #
# ---------------------------------------------- #

def test_linear_uniform_transitions_min():
    """Generate a uniform linear transition matrix with minimal states"""
    topology = _LinearTopology(n_states=1, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_linear_uniform_transitions_small():
    """Generate a uniform linear transition matrix with few states"""
    topology = _LinearTopology(n_states=2, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5],
        [0. , 1. ]
    ]))

def test_linear_uniform_transitions_many():
    """Generate a uniform linear transition matrix with many states"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.uniform_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.5, 0.5 , 0.        , 0.        , 0.        ],
        [0. , 0.5 , 0.5       , 0.        , 0.        ],
        [0. , 0.  , 0.5       , 0.5       , 0.        ],
        [0. , 0.  , 0.        , 0.5       , 0.5       ],
        [0. , 0.  , 0.        , 0.        , 1.        ]
    ]))

# --------------------------------------------- #
# _LinearTopology.random_transitions() #
# --------------------------------------------- #

def test_linear_random_transitions_min():
    """Generate a random linear transition matrix with minimal states"""
    topology = _LinearTopology(n_states=1, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [1.]
    ]))

def test_linear_random_transitions_small():
    """Generate a random linear transition matrix with few states"""
    topology = _LinearTopology(n_states=2, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.87426829, 0.12573171],
        [0.        , 1.        ]
    ]))

def test_linear_random_transitions_many():
    """Generate a random linear transition matrix with many states"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    assert_distribution(transitions)
    assert_equal(transitions, np.array([
        [0.9294571 , 0.0705429 , 0.        , 0.        , 0.        ],
        [0.        , 0.92269318, 0.07730682, 0.        , 0.        ],
        [0.        , 0.        , 0.86161736, 0.13838264, 0.        ],
        [0.        , 0.        , 0.        , 0.13863688, 0.86136312],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

# ----------------------------------------------- #
# _LinearTopology.validate_transitions() #
# ----------------------------------------------- #

def test_linear_validate_transitions_invalid():
    """Validate an invalid linear transition matrix"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = _ErgodicTopology(n_states=5, random_state=rng).random_transitions()
    with pytest.raises(ValueError) as e:
        topology.validate_transitions(transitions)
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_linear_validate_transitions_valid():
    """Validate a valid linear transition matrix"""
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    topology.validate_transitions(transitions)


import warnings, tqdm, tqdm.auto, numpy as np, types, pickle, marshal
from joblib import Parallel, delayed
from multiprocessing import cpu_count
from dtaidistance import dtw, dtw_ndim
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import LabelEncoder
from ...internals import _Validator

dtw_cc = None
try:
    from dtaidistance import dtw_cc
except ImportError:
    pass

class KNNClassifier:
    """A k-Nearest Neighbor classifier that uses dynamic time warping as a distance measure for comparing observation sequences of different length.

    Parameters
    ----------
    k: int > 0
        Number of neighbors.

    classes: array-like of str/numeric
        The complete set of possible classes/labels.

    weighting: 'uniform' or callable
        A callable that specifies how distance weighting should be performed.
        The callable should accept a :class:`numpy:numpy.ndarray` of DTW distances, apply an element-wise weighting transformation,
        then return an equally-sized :class:`numpy:numpy.ndarray` of weighted distances.

        If a `'uniform'` weighting is chosen, then the function ``lambda x: np.ones(x.size)`` is used, which weights all of the distances equally.

        If the callable is simple enough, it should be specified as a ``lambda``, but a function will also work.
        Examples of weighting functions are:

        - :math:`e^{-\\alpha x}`, specified by ``lambda x: np.exp(-alpha * x)`` for some positive :math:`\\alpha`/``alpha``,
        - :math:`\\frac{1}{x}`, specified by ``lambda x: 1 / x``.

        A good weighting function should *ideally* be defined at :math:`x=0` in the rare event that two observations are perfectly aligned and therefore have zero DTW distance.

        .. tip::
            It may be desirable to restrict DTW distances to a small range if you intend to use a weighting function.

            Using the :class:`~MinMaxScale` or :class:`~Standardize` preprocessing transformations to scale your features helps to ensure that DTW distances remain small.

    window: 0 ≤ float ≤ 1
        The width of the Sakoe-Chiba band global constraint as a fraction of the length of the longest of the two sequences.
        A larger constraint will speed up the DTW alignment by restricting the maximum temporal deviation from the diagonal of the DTW matrix,
        but too much constraint may lead to poor alignment.

        The default value of 1 corresponds to full DTW computation with no global constraint applied.

    use_c: bool
        Whether or not to use fast pure C compiled functions (from the `dtaidistance <https://github.com/wannesm/dtaidistance>`_ package) to perform the DTW computations.

        .. tip::
            If you set ``use_c = True`` and are receiving an error about a C library not being available, try reinstalling ``dtaidistance`` and disabling the cache:

            .. code-block:: console

                pip install -vvv --upgrade --no-cache-dir --force-reinstall dtaidistance

    independent: bool
        Whether or not to allow features to be warped independently from each other. See `here <https://www.cs.ucr.edu/~eamonn/Multi-Dimensional_DTW_Journal.pdf>`_ for a good overview of both approaches.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Attributes
    ----------
    k: int > 0
        The number of neighbors.

    weighting: callable
        The distance weighting function.

    window: 0 ≤ float ≤ 1
        The width of the Sakoe-Chiba band global constraint as a fraction of the length of the longest of the two sequences.

    use_c: bool
        Whether or not to use fast pure C compiled functions to perform the DTW computations.

    encoder: sklearn.preprocessing.LabelEncoder
        The label encoder fitted on the set of ``classes`` provided during instantiation.

    classes: numpy.ndarray (str/numeric)
        The complete set of possible classes/labels.
    """

    def __init__(self, k, classes, weighting='uniform', window=1., use_c=False, independent=False, random_state=None):
        self._val = _Validator()
        self._k = self._val.restricted_integer(
            k, lambda x: x > 0, desc='number of neighbors', expected='greater than zero')
        self._window = float(window) if window in (0, 1) else self._val.restricted_float(
            window, lambda x: 0. <= x <= 1., desc='Sakoe-Chiba band width (fraction)', expected='between zero and one')
        self._random_state = self._val.random_state(random_state)

        self._val.iterable(classes, 'classes')
        self._val.string_or_numeric(classes[0], 'each class')
        if all(isinstance(label, type(classes[0])) for label in classes[1:]):
            self._encoder = LabelEncoder().fit(classes)
        else:
            raise TypeError('Expected all classes to be of the same type')

        if weighting == 'uniform':
            self._weighting = lambda x: np.ones(x.size)
        else:
            self._val.func(weighting, 'distance weighting function')
            try:
                if isinstance(weighting(np.ones(5)), np.ndarray):
                    self._weighting = weighting
                else:
                    raise TypeError('Expected weighting function to accept a numpy.ndarray and return an equally-sized numpy.ndarray')
            except:
                raise TypeError('Expected weighting function to accept a numpy.ndarray and return an equally-sized numpy.ndarray')

        self._use_c = self._val.boolean(use_c, desc='whether or not to use fast pure C compiled functions')
        if self._use_c and (dtw_cc is None):
            warnings.warn('DTAIDistance C library not available – using Python implementation', ImportWarning)
            self._use_c = False

        self._independent = self._val.boolean(independent, 'independent')
        self._dtw = self._dtwi if independent else self._dtwd

    def fit(self, X, y):
        """Fits the classifier by adding labeled training observation sequences.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            A list of multiple observation sequences.

        y: array-like of str/numeric
            An iterable of labels for the observation sequences.
        """
        X, y = self._val.observation_sequences_and_labels(X, y)
        self._X, self._y = X, self._encoder.transform(y)
        self._n_features = X[0].shape[1]

    def predict(self, X, original_labels=True, verbose=True, n_jobs=1):
        """Predicts the label for an observation sequence (or multiple sequences).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        original_labels: bool
            Whether to inverse-transform the labels to their original encoding.

        verbose: bool
            Whether to display a progress bar or not.

            .. note::
                If both ``verbose=True`` and ``n_jobs > 1``, then the progress bars for each process
                are always displayed in the console, regardless of where you are running this function from
                (e.g. a Jupyter notebook).

        n_jobs: int > 0 or -1
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        prediction(s): str/numeric or :class:`numpy:numpy.ndarray` (str/numeric)
            The predicted label(s) for the observation sequence(s).

            If ``original_labels`` is true, then the returned labels are
            inverse-transformed into their original encoding.
        """
        try:
            (self._X, self._y)
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted before predictions are made')

        X = self._val.observation_sequences(X, allow_single=True)
        self._val.boolean(original_labels, desc='original_labels')
        self._val.boolean(verbose, desc='verbose')
        self._val.restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        if isinstance(X, np.ndarray):
            distances = np.array([self._dtw(X, x) for x in tqdm.auto.tqdm(self._X, desc='Calculating distances', disable=not(verbose))])
            return self._output(self._find_nearest(distances), original_labels)
        else:
            n_jobs = min(cpu_count() if n_jobs == -1 else n_jobs, len(X))
            X_chunks = [list(chunk) for chunk in np.array_split(np.array(X, dtype=object), n_jobs)]
            labels = Parallel(n_jobs=n_jobs)(delayed(self._chunk_predict)(i+1, chunk, verbose) for i, chunk in enumerate(X_chunks))
            return self._output(np.concatenate(labels), original_labels) # Flatten the resulting array

    def evaluate(self, X, y, verbose=True, n_jobs=1):
        """Evaluates the performance of the classifier on a batch of observation sequences and their labels.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            A list of multiple observation sequences.

        y: array-like of str/numeric
            An iterable of labels for the observation sequences.

        verbose: bool
            Whether to display a progress bar for predictions or not.

        n_jobs: int > 0 or -1
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        accuracy: float
            The categorical accuracy of the classifier on the observation sequences.

        confusion: :class:`numpy:numpy.ndarray` (int)
            The confusion matrix representing the discrepancy between predicted and actual labels.
        """
        X, y = self._val.observation_sequences_and_labels(X, y)
        self._val.boolean(verbose, desc='verbose')
        predictions = self.predict(X, original_labels=False, verbose=verbose, n_jobs=n_jobs)
        cm = confusion_matrix(self._encoder.transform(y), predictions, labels=self._encoder.transform(self._encoder.classes_))
        return np.sum(np.diag(cm)) / np.sum(cm), cm

    def save(self, path):
        """Serializes the :class:`KNNClassifier` object by pickling its hyper-parameters, variables and training data.

            .. warning::
                As :math:`k`-NN classifier must look through each training example during prediction,
                saving the classifier simply copies all of the training observation sequences and labels.

        Parameters
        ----------
        path: str
            File path (usually with `.pkl` extension) to store the serialized :class:`KNNClassifier` object.
        """
        try:
            (self._X, self._y)
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted before it can be saved')

        # Pickle the necessary hyper-parameters, variables and data
        with open(path, 'wb') as file:
            pickle.dump({
                'k': self._k,
                'classes': self._encoder.classes_,
                # Serialize the weighting function into a byte-string
                'weighting': marshal.dumps((self._weighting.__code__, self._weighting.__name__)),
                'window': self._window,
                'use_c': self._use_c,
                'independent': self._independent,
                'random_state': self._random_state,
                'X': self._X,
                'y': self._y,
                'n_features': self._n_features
            }, file)

    @classmethod
    def load(cls, path):
        """Deserializes a :class:`KNNClassifier` object which was serialized with the :meth:`save` function.

        Parameters
        ----------
        path: str
            File path of the serialized data generated by the :meth:`save` method.

        Returns
        -------
        deserialized: :class:`KNNClassifier`
            The deserialized DTW :math:`k`-NN classifier object.
        """
        with open(path, 'rb') as file:
            data = pickle.load(file)

            # Check deserialized object dictionary and keys
            keys = set(('k', 'classes', 'weighting', 'window', 'use_c', 'independent', 'random_state', 'X', 'y', 'n_features'))
            if not isinstance(data, dict):
                raise TypeError('Expected deserialized object to be a dictionary - make sure the object was serialized with the save() function')
            else:
                if len(set(keys) - set(data.keys())) != 0:
                    raise ValueError('Missing keys in deserialized object dictionary – make sure the object was serialized with the save() function')

            # Deserialize the weighting function
            weighting, name = marshal.loads(data['weighting'])
            weighting = types.FunctionType(weighting, globals(), name)

            # Instantiate a new KNNClassifier with the same hyper-parameters
            clf = cls(
                k=data['k'],
                classes=data['classes'],
                weighting=weighting,
                window=data['window'],
                use_c=data['use_c'],
                independent=data['independent'],
                random_state=data['random_state']
            )

            # Load the data directly
            clf._X, clf._y = data['X'], data['y']
            clf._n_features = data['n_features']

            return clf

    def _dtw_1d(self, a, b, window): # Requires fit
        """Computes the DTW distance between two univariate sequences."""
        return dtw.distance(a, b, use_c=self._use_c, window=window)

    def _dtwi(self, A, B): # Requires fit
        """Computes the multivariate DTW distance as the sum of the pairwise per-feature DTW distances, allowing each feature to be warped independently."""
        window = max(1, int(self._window * max(len(A), len(B))))
        return np.sum([self._dtw_1d(A[:, i], B[:, i], window=window) for i in range(self._n_features)])

    def _dtwd(self, A, B): # Requires fit
        """Computes the multivariate DTW distance so that the warping of the features depends on each other, by modifying the local distance measure."""
        window = max(1, int(self._window * max(len(A), len(B))))
        return dtw_ndim.distance(A, B, use_c=self._use_c, window=window)

    def _argmax(self, a):
        """Same as numpy.argmax but returns all occurrences of the maximum, and is O(n) instead of O(2n).
        From: https://stackoverflow.com/a/58652335
        """
        all_, max_ = [0], a[0]
        for i in range(1, len(a)):
            if a[i] > max_:
                all_, max_ = [i], a[i]
            elif a[i] == max_:
                all_.append(i)
        return np.array(all_)

    def _find_nearest(self, distances): # Requires fit
        """Returns the mode label of the k nearest neighbors.
        Vectorization from: https://stackoverflow.com/a/49239335
        """
        # Find the indices, labels and distances of the k-nearest neighbours
        idx = np.argpartition(distances, self._k)[:self._k]
        nearest_labels = self._y[idx]
        nearest_distances = self._weighting(distances[idx])
        # Combine labels and distances into one array and sort by label
        labels_distances = np.vstack((nearest_labels, nearest_distances))
        labels_distances = labels_distances[:, labels_distances[0, :].argsort()]
        # Find indices where the label changes
        i = np.nonzero(np.diff(labels_distances[0, :]))[0] + 1
        i = np.insert(i, 0, 0)
        # Add-reduce weighted distances within each label group (ordered by label)
        label_scores = np.add.reduceat(labels_distances[1, :], i)
        # Find the mode labels (set of labels with labels scores equal to the maximum)
        max_labels = nearest_labels[self._argmax(label_scores)]
        # Randomly pick from the set of labels with the maximum label score
        return self._random_state.choice(max_labels)

    def _chunk_predict(self, process, chunk, verbose): # Requires fit
        """Makes predictions for a chunk of the observation sequences, for a given subprocess."""
        labels = np.zeros(len(chunk), dtype=int)
        for i, sequence in enumerate(tqdm.auto.tqdm(chunk,
            desc='Classifying examples (process {})'.format(process),
            disable=not(verbose), position=process-1)
        ):
            distances = np.array([self._dtw(sequence, x) for x in self._X])
            labels[i] = self._find_nearest(distances)
        return labels

    def _output(self, out, original_labels):
        """Inverse-transforms the labels if necessary, and returns them."""
        if isinstance(out, np.ndarray):
            return self._encoder.inverse_transform(out) if original_labels else out
        else:
            return self._encoder.inverse_transform([out]).item() if original_labels else out

    @property
    def k(self):
        return self._k

    @property
    def weighting(self):
        return self._weighting

    @property
    def window(self):
        return self._window

    @property
    def use_c(self):
        return self._use_c

    @property
    def encoder(self):
        return self._encoder

    @property
    def classes(self):
        return self._encoder.classes_

    @property
    def X(self):
        try:
            return self._X
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted first')

    @property
    def y(self):
        try:
            return self._y
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted first')

    def __repr__(self):
        module = self.__class__.__module__
        out = '{}{}('.format('' if module == '__main__' else '{}.'.format(module), self.__class__.__name__)
        attrs = [
            ('k', repr(self._k)),
            ('window', repr(self._window)),
            ('use_c', repr(self._use_c)),
            ('independent', repr(self._independent)),
            ('classes', repr(list(self._encoder.classes_)))
        ]
        try:
            (self._X, self._y)
            attrs.extend([('X', '[...]'), ('y', 'array([...])')])
        except AttributeError:
            pass
        return out + ', '.join('{}={}'.format(name, val) for name, val in attrs) + ')'


import numpy as np
from warnings import warn
from .topology import _Topology

class _ErgodicTopology(_Topology):
    """Represents the topology for an ergodic HMM, imposing non-zero probabilities in the transition matrix.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
            to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        return np.ones((self._n_states, self._n_states)) / self._n_states

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution - according to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The random transition matrix of shape `(n_states, n_states)`.
        """
        return self._random_state.dirichlet(np.ones(self._n_states), size=self._n_states)

    def validate_transitions(self, transitions: np.ndarray) -> None:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray (float)
            The transition matrix to validate.
        """
        super().validate_transitions(transitions)
        if not np.all(transitions > 0):
            warn('Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned')


import numpy as np
from copy import copy
from tqdm.auto import tqdm
from ..internals import _Validator

__all__ = ['Transform', 'Equalize', 'TrimZeros', 'MinMaxScale', 'Center', 'Standardize', 'Downsample', 'Filter']

class Transform:
    """Base class representing a single transformation."""

    def __init__(self):
        self._val = _Validator()

    def _describe(self):
        """Description of the transformation.

        Returns
        -------
        description: str
            The description of the transformation.
        """
        raise NotImplementedError

    def transform(self, X, verbose=False):
        """Applies the transformation.

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether or not to display a progress bar when applying transformations.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float) or list of :class:`numpy:numpy.ndarray` (float)
            The transformed input observation sequence(s).
        """
        raise NotImplementedError

    def __call__(self, X, verbose=False):
        """Alias of the :meth:`transform` method."""
        return self.transform(X, verbose)

    def _apply(self, transform, X, verbose):
        """Applies the transformation to the observation sequence(s).

        Parameters
        ----------
        transform: callable
            The transformation to apply.

        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        Returns
        -------
        transformed: numpy.ndarray (float) or list of numpy.ndarray (float)
            The transformed input observation sequence(s).
        """
        X = self._val.observation_sequences(X, allow_single=True)
        verbose = self._val.boolean(verbose, 'verbose')

        def apply_transform():
            if isinstance(X, np.ndarray):
                return transform(copy(X))
            else:
                return [transform(x) for x in tqdm(copy(X), desc=self._describe(), disable=not(verbose))]

        if self._is_fitted():
            return apply_transform()
        else:
            try:
                self.fit(X)
                return apply_transform()
            except:
                raise
            finally:
                self._unfit()

    def fit(self, X):
        """Fit the transformation on the provided observation sequence(s) (without transforming them).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.
        """
        self._val.observation_sequences(X, allow_single=True)

    def _unfit(self):
        """Unfit the transformation by resetting the parameters to their default settings."""
        pass

    def _is_fitted(self):
        """Check whether or not the transformation is fitted on some observation sequence(s).

        Returns
        -------
        fitted: bool
            Whether or not the transformation is fitted.
        """
        return False

    def fit_transform(self, X, verbose=False):
        """Fit the transformation with the provided observation sequence(s) and transform them.

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether or not to display a progress bar when fitting and applying transformations.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float) or list of :class:`numpy:numpy.ndarray` (float)
            The transformed input observation sequence(s).
        """
        self.fit(X)
        return self.transform(X, verbose)

class Equalize(Transform):
    """Equalize all observation sequence lengths by padding or trimming zeros."""

    def __init__(self):
        super().__init__()
        self.length = None

    def fit(self, X):
        X = self._val.observation_sequences(X, allow_single=True)
        self.length = max(len(x) for x in X) if isinstance(X, list) else len(X)

    def _unfit(self):
        self.length = None

    def _is_fitted(self):
        return self.length is not None

    def _describe(self):
        return 'Equalize sequence lengths'

    def transform(self, X, verbose=False):
        def equalize(x):
            T, D = x.shape
            return np.vstack((x, np.zeros((self.length - T, D)))) if T <= self.length else x[:self.length]
        return self._apply(equalize, X, verbose)

class TrimZeros(Transform):
    """Trim zero-observations from the input observation sequence(s).

    Examples
    --------
    >>> # Create some sample data
    >>> z = np.zeros((4, 3))
    >>> x = lambda i: np.vstack((z, np.random.random((10 * i, 3)), z))
    >>> X = [x(i) for i in range(1, 4)]
    >>> # Zero-trim the data
    >>> X = TrimZeros()(X)
    """

    def _describe(self):
        return 'Remove zero-observations'

    def transform(self, X, verbose=False):
        def trim_zeros(x):
            return x[~np.all(x == 0, axis=1)]
        return self._apply(trim_zeros, X, verbose)

class MinMaxScale(Transform):
    """Scales the observation sequence features to each be within a provided range.

    Parameters
    ----------
    scale: tuple(int/float, int/float)
        The range of the transformed observation sequence features.

    independent: bool
        Whether to independently compute the minimum and maximum to scale each observation sequence.
    """

    def __init__(self, scale=(0., 1.), independent=True):
        super().__init__()
        if not isinstance(scale, tuple):
            raise TypeError('Expected scaling range to be a tuple')
        if not all(isinstance(val, (int, float)) for val in scale):
            raise TypeError('Expected upper and lower bounds of scaling range to be floats')
        if not scale[0] < scale[1]:
            raise ValueError('Expected lower bound of scaling range to be less than the upper bound')
        self.scale = scale
        self.independent = self._val.boolean(independent, 'independent')
        if not self.independent:
            self.min, self.max = None, None

    def fit(self, X):
        X = self._val.observation_sequences(X, allow_single=True)
        if not self.independent:
            X_concat = np.vstack(X) if isinstance(X, list) else X
            self.min, self.max = X_concat.min(axis=0), X_concat.max(axis=0)

    def _unfit(self):
        if not self.independent:
            self.min, self.max = None, None

    def _is_fitted(self):
        return False if self.independent else (self.min is not None) and (self.max is not None)

    def _describe(self):
        if not self.independent:
            return 'Min-max scaling into range {}'.format(self.scale)
        else:
            return 'Min-max scaling into range {} (independent)'.format(self.scale)

    def transform(self, X, verbose=False):
        if not self.independent:
            def min_max_scale(x):
                min_, max_ = self.scale
                scale = (max_ - min_) / (self.max - self.min)
                return scale * x + min_ - self.min * scale
        else:
            def min_max_scale(x):
                min_, max_ = self.scale
                scale = (max_ - min_) / (x.max(axis=0) - x.min(axis=0))
                return scale * x + min_ - x.min(axis=0) * scale
        return self._apply(min_max_scale, X, verbose)

class Center(Transform):
    """Centers the observation sequence features around their means. Results in zero-mean features.

    Parameters
    ----------
    independent: bool
        Whether to independently compute the mean to scale each observation sequence.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Center the data
    >>> X = Center()(X)
    """

    def __init__(self, independent=True):
        super().__init__()
        self.independent = self._val.boolean(independent, 'independent')
        if not self.independent:
            self.mean = None

    def fit(self, X):
        X = self._val.observation_sequences(X, allow_single=True)
        if not self.independent:
            X_concat = np.vstack(X) if isinstance(X, list) else X
            self.mean = X_concat.mean(axis=0)

    def _unfit(self):
        if not self.independent:
            self.mean = None

    def _is_fitted(self):
        return False if self.independent else (self.mean is not None)

    def _describe(self):
        if not self.independent:
            return 'Centering around mean (zero mean)'
        else:
            return 'Centering around mean (zero mean) (independent)'

    def transform(self, X, verbose=False):
        if not self.independent:
            def center(x):
                return x - self.mean
        else:
            def center(x):
                return x - x.mean(axis=0)
        return self._apply(center, X, verbose)

class Standardize(Transform):
    """Centers the observation sequence features around their means, then scales them by their deviations.
    Results in zero-mean, unit-variance features.

    Parameters
    ----------
    independent: bool
        Whether to independently compute the mean and standard deviation to scale each observation sequence.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Standardize the data
    >>> X = Standardize()(X)
    """

    def __init__(self, independent=True):
        super().__init__()
        self.independent = self._val.boolean(independent, 'independent')
        if not self.independent:
            self.mean, self.std = None, None

    def fit(self, X):
        X = self._val.observation_sequences(X, allow_single=True)
        if not self.independent:
            X_concat = np.vstack(X) if isinstance(X, list) else X
            self.mean, self.std = X_concat.mean(axis=0), X_concat.std(axis=0)

    def _unfit(self):
        if not self.independent:
            self.mean, self.std = None, None

    def _is_fitted(self):
        return False if self.independent else (self.mean is not None) and (self.std is not None)

    def _describe(self):
        if not self.independent:
            return 'Standard scaling (zero mean, unit variance)'
        else:
            return 'Standard scaling (zero mean, unit variance) (independent)'

    def transform(self, X, verbose=False):
        if not self.independent:
            def standardize(x):
                return (x - self.mean) / self.std
        else:
            def standardize(x):
                return (x - x.mean(axis=0)) / x.std(axis=0)
        return self._apply(standardize, X, verbose)

class Downsample(Transform):
    """Downsamples an observation sequence (or multiple sequences) by either:

    - Decimating the next :math:`n-1` observations
    - Averaging the current observation with the next :math:`n-1` observations

    Parameters
    ----------
    factor: int > 0
        Downsample factor.

    method: {'decimate', 'mean'}
        The downsampling method.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Downsample the data with downsample factor 5 and decimation
    >>> X = Downsample(factor=5, method='decimate')(X)
    """

    def __init__(self, factor, method='decimate'):
        super().__init__()
        self.factor = self._val.restricted_integer(factor, lambda x: x > 0, desc='downsample factor', expected='positive')
        self.method = self._val.one_of(method, ['decimate', 'mean'], desc='downsampling method')

    def _describe(self):
        method = 'Decimation' if self.method == 'decimate' else 'Mean'
        return '{} downsampling with factor {}'.format(method, self.factor)

    def transform(self, X, verbose=False):
        X = self._val.observation_sequences(X, allow_single=True)

        if isinstance(X, np.ndarray):
            self._val.restricted_integer(self.factor, lambda x: x <= len(X),
                desc='downsample factor', expected='no greater than the number of frames')
        else:
            self._val.restricted_integer(self.factor, lambda x: x <= min(len(x) for x in X),
                desc='downsample factor', expected='no greater than the number of frames in the shortest sequence')

        def downsample(x):
            N, D = x.shape
            if self.method == 'decimate':
                return np.delete(x, [i for i in range(N) if i % self.factor != 0], 0)
            elif self.method == 'mean':
                r = len(x) % self.factor
                xn, xr = (x, None) if r == 0 else (x[:-r], x[-r:])
                dxn = xn.T.reshape(-1, self.factor).mean(axis=1).reshape(D, -1).T
                return dxn if xr is None else np.vstack((dxn, xr.mean(axis=0)))

        return self._apply(downsample, X, verbose)

class Filter(Transform):
    """Applies a median or mean filter to the input observation sequence(s).

    Parameters
    ----------
    window_size: int > 0
        The size of the filtering window.

    method: {'median', 'mean'}
        The filtering method.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((10 * i, 3)) for i in range(1, 4)]
    >>> # Filter the data with window size 5 and median filtering
    >>> X = Filter(window_size=5, method='median')(X)
    """

    def __init__(self, window_size, method='median'):
        super().__init__()
        self.window_size = self._val.restricted_integer(window_size, lambda x: x > 0, desc='window size', expected='positive')
        self.method = self._val.one_of(method, ['median', 'mean'], desc='filtering method')

    def _describe(self):
        return '{} filtering with window-size {}'.format(self.method.capitalize(), self.window_size)

    def transform(self, X, verbose=False):
        X = self._val.observation_sequences(X, allow_single=True)

        if isinstance(X, np.ndarray):
            self._val.restricted_integer(self.window_size, lambda x: x <= len(X),
                desc='window size', expected='no greater than the number of frames')
        else:
            self._val.restricted_integer(self.window_size, lambda x: x <= min(len(x) for x in X),
                desc='window size', expected='no greater than the number of frames in the shortest sequence')

        def filter_(x):
            measure = np.median if self.method == 'median' else np.mean
            filtered = []
            right = self.window_size // 2
            left = (self.window_size - 1) - right
            for i in range(len(x)):
                l, m, r = x[((i - left) * (left < i)):i], x[i], x[(i + 1):(i + 1 + right)]
                filtered.append(measure(np.vstack((l, m, r)), axis=0))
            return np.array(filtered)

        return self._apply(filter_, X, verbose)


import numpy as np
from .topology import _Topology

class _LinearTopology(_Topology):
    """Represents the topology for a linear HMM.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
            to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        transitions = np.zeros((self._n_states, self._n_states))
        for i, row in enumerate(transitions[:-1]):
            row[i:(i+2)] = np.ones(2) / 2
        transitions[self._n_states - 1][self._n_states - 1] = 1
        return transitions

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution, according to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The random transition matrix of shape `(n_states, n_states)`.
        """
        transitions = np.zeros((self._n_states, self._n_states))
        for i, row in enumerate(transitions[:-1]):
            row[i:(i+2)] = self._random_state.dirichlet(np.ones(2))
        transitions[self._n_states - 1][self._n_states - 1] = 1
        return transitions

    def validate_transitions(self, transitions: np.ndarray) -> None:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray (float)
            The transition matrix to validate.
        """
        super().validate_transitions(transitions)
        if not np.allclose(transitions, np.triu(transitions)):
            raise ValueError('Left-right transition matrix must be upper-triangular')
        if not np.allclose(transitions, np.diag(np.diag(transitions)) + np.diag(np.diag(transitions, k=1), k=1)):
            raise ValueError('Linear transition matrix must only consist of a diagonal and upper diagonal')


import os, pickle, pytest, warnings, os, numpy as np, hmmlearn.hmm
from copy import deepcopy
from sequentia.classifiers import GMMHMM, HMMClassifier, _ErgodicTopology
from ....support import assert_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# Set of possible labels
labels = ['c{}'.format(i) for i in range(5)]

# Create and fit some sample HMMs
hmms = []
for i, label in enumerate(labels):
    hmm = GMMHMM(label=label, n_states=(i + 3), random_state=rng)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit([np.arange((i + j * 20) * 30).reshape(-1, 3) for j in range(2, 5)])
    hmms.append(hmm)

# Create some sample test data and labels
X = [np.arange((i + 2 * 20) * 30).reshape(-1, 3) for i in range(2, 5)]
Y = ['c0', 'c1', 'c1']
x, y = X[0], 'c1'

# Fit a classifier
hmm_clf = HMMClassifier()
hmm_clf.fit(hmms)

# =================== #
# HMMClassifier.fit() #
# =================== #

def test_fit_list():
    """Fit the classifier using a list of HMMs"""
    clf = HMMClassifier()
    clf.fit(hmms)
    assert clf._models == hmms

def test_fit_list_empty():
    """Fit the classifier using an empty list"""
    clf = HMMClassifier()
    with pytest.raises(RuntimeError) as e:
        clf.fit([])
    assert str(e.value) == 'The classifier must be fitted with at least one HMM'

def test_fit_list_invalid():
    """Fit the classifier using a list of invalid types"""
    clf = HMMClassifier()
    with pytest.raises(TypeError) as e:
        clf.fit(['a', 'b'])
    assert str(e.value) == 'Expected all models to be GMMHMM objects'

# ======================= #
# HMMClassifier.predict() #
# ======================= #

def test_predict_single_frequency_prior():
    """Predict a single observation sequence with a frequency prior"""
    prediction = hmm_clf.predict(x, prior='frequency', return_scores=False, original_labels=False)
    assert prediction == 4

def test_predict_single_uniform_prior():
    """Predict a single observation sequence with a uniform prior"""
    prediction = hmm_clf.predict(x, prior='uniform', return_scores=False, original_labels=False)
    assert prediction == 4

def test_predict_single_custom_prior():
    """Predict a single observation sequence with a custom prior"""
    prediction = hmm_clf.predict(x, prior=([1-4e-10]+[1e-10]*4), return_scores=False, original_labels=False)
    assert prediction == 0

def test_predict_single_return_scores():
    """Predict a single observation sequence and return the transformed label, with the un-normalized posterior scores"""
    prediction = hmm_clf.predict(x, prior='frequency', return_scores=True, original_labels=False)
    assert isinstance(prediction, tuple)
    assert prediction[0] == 4
    assert_equal(prediction[1], np.array([
        -1224.7910008 , -1302.74962802, -1269.35306368, -1230.59148179, -1222.21963107
    ]))

def test_predict_single_original_labels():
    """Predict a single observation sequence and return the original label, without the un-normalized posterior scores"""
    prediction = hmm_clf.predict(x, prior='uniform', return_scores=False, original_labels=True)
    assert prediction == 'c4'

def test_predict_single_return_scores_original_labels():
    """Predict a single observation sequence and return the original label, with the un-normalized posterior scores"""
    prediction = hmm_clf.predict(x, prior='frequency', return_scores=True, original_labels=True)
    assert isinstance(prediction, tuple)
    assert prediction[0] == 'c4'
    assert_equal(prediction[1], np.array([
        -1224.7910008 , -1302.74962802, -1269.35306368, -1230.59148179, -1222.21963107
    ]))

def test_predict_multiple_frequency_prior():
    """Predict multiple observation sequences with a frequency prior"""
    predictions = hmm_clf.predict(X, prior='frequency', return_scores=False, original_labels=False)
    assert_equal(predictions, np.array([4, 4, 0]))

def test_predict_multiple_uniform_prior():
    """Predict multiple observation sequences with a uniform prior"""
    predictions = hmm_clf.predict(X, prior='uniform', return_scores=False, original_labels=False)
    assert_equal(predictions, np.array([4, 4, 0]))

def test_predict_multiple_custom_prior():
    """Predict multiple observation sequences with a custom prior"""
    predictions = hmm_clf.predict(X, prior=([1-4e-10]+[1e-10]*4), return_scores=False, original_labels=False)
    assert_equal(predictions, np.array([0, 0, 0]))

def test_predict_multiple_return_scores():
    """Predict multiple observation sequences and return the transformed labels, with the un-normalized posterior scores"""
    predictions = hmm_clf.predict(X, prior='frequency', return_scores=True, original_labels=False)
    assert isinstance(predictions, tuple)
    assert_equal(predictions[0], np.array([4, 4, 0]))
    assert_equal(predictions[1], np.array([
        [-1224.7910008 , -1302.74962802, -1269.35306368, -1230.59148179, -1222.21963107],
        [-1253.13158379, -1331.2088869 , -1303.04688636, -1259.4763248 , -1251.90302358],
        [-1281.50581789, -1359.72306473, -1335.76859787, -1288.46683118, -1281.66788896]
    ]))

def test_predict_multiple_original_labels():
    """Predict multiple observation sequences and return the original labels, without the un-normalized posterior scores"""
    predictions = hmm_clf.predict(X, prior='frequency', return_scores=False, original_labels=True)
    assert all(np.equal(predictions.astype(object), np.array(['c4', 'c4', 'c0'], dtype=object)))

def test_predict_multiple_return_scores_original_labels():
    """Predict multiple observation sequences and return the original labels, with the un-normalized posterior scores"""
    predictions = hmm_clf.predict(X, prior='frequency', return_scores=True, original_labels=True)
    assert isinstance(predictions, tuple)
    assert all(np.equal(predictions[0].astype(object), np.array(['c4', 'c4', 'c0'], dtype=object)))
    assert_equal(predictions[1], np.array([
        [-1224.7910008 , -1302.74962802, -1269.35306368, -1230.59148179, -1222.21963107],
        [-1253.13158379, -1331.2088869 , -1303.04688636, -1259.4763248 , -1251.90302358],
        [-1281.50581789, -1359.72306473, -1335.76859787, -1288.46683118, -1281.66788896]
    ]))

# ======================== #
# HMMClassifier.evaluate() #
# ======================== #

def test_evaluate():
    """Evaluate performance on some observation sequences and labels"""
    acc, cm = hmm_clf.evaluate(X, Y, prior='frequency')
    assert acc == 0.0
    assert_equal(cm, np.array([
        [0, 0, 0, 0, 1],
        [1, 0, 0, 0, 1],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0]
    ]))

# ===================================== #
# HMMClassifier (serialize/deserialize) #
# ===================================== #

def test_serialization():
    """Serialize and deserialize the classifier"""
    model_file = 'test.pkl'
    try:
        with open(model_file, 'wb') as file:
            pickle.dump(hmm_clf, file)
        with open(model_file, 'rb') as file:
            clf = pickle.load(file)
            predictions = clf.predict(X, prior='frequency', return_scores=True, original_labels=True)
            assert isinstance(predictions, tuple)
            assert all(np.equal(predictions[0].astype(object), np.array(['c4', 'c4', 'c0'], dtype=object)))
            assert_equal(predictions[1], np.array([
                [-1224.7910008 , -1302.74962802, -1269.35306368, -1230.59148179, -1222.21963107],
                [-1253.13158379, -1331.2088869 , -1303.04688636, -1259.4763248 , -1251.90302358],
                [-1281.50581789, -1359.72306473, -1335.76859787, -1288.46683118, -1281.66788896]
            ]))
    finally:
        if os.path.exists(model_file):
            os.remove(model_file)

# ==================== #
# HMMClassifier.save() #
# ==================== #

def test_save_unfitted():
    """Save an unfitted HMMClassifier object."""
    try:
        with pytest.raises(RuntimeError) as e:
            HMMClassifier().save('test.pkl')
        assert str(e.value) == 'The classifier needs to be fitted before it can be saved'
    finally:
        if os.path.exists('test.pkl'):
            os.remove('test.pkl')

def test_save_fitted():
    """Save a fitted HMMClassifier object."""
    try:
        hmm_clf.save('test.pkl')
        assert os.path.isfile('test.pkl')
    finally:
        os.remove('test.pkl')

# ==================== #
# HMMClassifier.load() #
# ==================== #

def test_load_invalid_format():
    """Load a HMMClassifier from an illegally formatted file"""
    try:
        with open('test', 'w') as f:
            f.write('illegal')
        with pytest.raises(pickle.UnpicklingError) as e:
            HMMClassifier.load('test')
    finally:
        os.remove('test')

def test_load_valid():
    """Load a serialized HMMClassifier"""
    try:
        hmm_clf.save('test.pkl')
        clf = HMMClassifier.load('test.pkl')
        # Check that all fields are still the same
        assert isinstance(clf, HMMClassifier)
        assert all(isinstance(model, GMMHMM) for model in clf._models)
        assert [model.label for model in clf._models] == labels
        assert list(clf._encoder.classes_) == labels
    finally:
        os.remove('test.pkl')


import pytest, warnings, os, numpy as np, hmmlearn.hmm
from copy import deepcopy
from sequentia.classifiers import GMMHMM, _LeftRightTopology, _ErgodicTopology, _LinearTopology
from ....support import assert_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# Create some sample data
X = [rng.random((10 * i, 3)) for i in range(1, 4)]
x = rng.random((15, 3))

# Unparameterized HMMs
hmm_lr = GMMHMM(label='c1', n_states=5, topology='left-right', random_state=rng)
hmm_e = GMMHMM(label='c1', n_states=5, topology='ergodic', random_state=rng)
hmm_lin = GMMHMM(label='c1', n_states=5, topology='linear', random_state=rng)

# ======================================================== #
# GMMHMM.set_uniform_initial() + GMMHMM.initial (property) #
# ======================================================== #

def test_left_right_uniform_initial():
    """Uniform initial state distribution for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

def test_ergodic_uniform_initial():
    """Uniform initial state distribution for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

def test_linear_uniform_initial():
    """Uniform initial state distribution for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

# ======================================================= #
# GMMHMM.set_random_initial() + GMMHMM.initial (property) #
# ======================================================= #

def test_left_right_random_initial():
    """Random initial state distribution for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))

def test_ergodic_random_initial():
    """Random initial state distribution for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))

def test_linear_random_initial():
    """Random initial state distribution for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))

# ================================================================ #
# GMMHMM.set_uniform_transitions() + GMMHMM.transitions (property) #
# ================================================================ #

def test_left_right_uniform_transitions():
    """Uniform transition matrix for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.2, 0.2 , 0.2       , 0.2       , 0.2       ],
        [0. , 0.25, 0.25      , 0.25      , 0.25      ],
        [0. , 0.  , 0.33333333, 0.33333333, 0.33333333],
        [0. , 0.  , 0.        , 0.5       , 0.5       ],
        [0. , 0.  , 0.        , 0.        , 1.        ]
    ]))

def test_ergodic_uniform_transitions():
    """Uniform transition matrix for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))

def test_linear_uniform_transitions():
    """Uniform transition matrix for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.5, 0.5, 0. , 0. , 0. ],
        [0. , 0.5, 0.5, 0. , 0. ],
        [0. , 0. , 0.5, 0.5, 0. ],
        [0. , 0. , 0. , 0.5, 0.5],
        [0. , 0. , 0. , 0. , 1. ]
    ]))

# =============================================================== #
# GMMHMM.set_random_transitions() + GMMHMM.transitions (property) #
# =============================================================== #

def test_left_right_random_transitions():
    """Random transition matrix for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597],
        [0.        , 0.20580715, 0.5280311 , 0.06521685, 0.20094491],
        [0.        , 0.        , 0.33761567, 0.2333124 , 0.42907193],
        [0.        , 0.        , 0.        , 0.36824778, 0.63175222],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

def test_ergodic_random_transitions():
    """Random transition matrix for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597],
        [0.19252534, 0.15767581, 0.47989976, 0.01708551, 0.15281357],
        [0.19375092, 0.16425506, 0.21828034, 0.11397708, 0.30973661],
        [0.46906977, 0.02941216, 0.17137502, 0.0333193 , 0.29682374],
        [0.21312406, 0.35221103, 0.08556524, 0.06613143, 0.28296824]
    ]))

def test_linear_random_transitions():
    """Random transition matrix for a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.72413873, 0.27586127, 0.        , 0.        , 0.        ],
        [0.        , 0.07615418, 0.92384582, 0.        , 0.        ],
        [0.        , 0.        , 0.81752797, 0.18247203, 0.        ],
        [0.        , 0.        , 0.        , 0.24730529, 0.75269471],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

# ============================ #
# GMMHMM.fit() + GMMHMM.n_seqs #
# ============================ #

def test_fit_without_initial_and_transition():
    """Fitting before setting the initial state distribution and transition matrix"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.fit(X)
    assert str(e.value) == 'Must specify initial state distribution and transitions before the HMM can be fitted'

def test_fit_without_transitions():
    """Fitting before setting the initial transition matrix"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    with pytest.raises(AttributeError) as e:
        hmm.fit(X)
    assert str(e.value) == 'Must specify initial state distribution and transitions before the HMM can be fitted'

def test_fit_without_initial():
    """Fitting before setting the initial state distribution"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.fit(X)
    assert str(e.value) == 'Must specify initial state distribution and transitions before the HMM can be fitted'

def test_fit_sets_internals():
    """Check that fitting sets internal attributes"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    hmm.fit(X)
    assert hmm.n_seqs == 3
    assert isinstance(hmm._model, hmmlearn.hmm.GMMHMM)

def test_left_right_fit_updates_uniform_initial():
    """Check fitting with a uniform initial state distribution of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    before = hmm.initial
    hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_left_right_fit_updates_random_initial():
    """Check fitting with a random initial state distribution of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))
    before = hmm.initial
    hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_left_right_fit_updates_uniform_transitions():
    """Check fitting with a uniform transition matrix of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.transitions
    assert_equal(hmm.transitions, np.array([
        [0.2       , 0.2       , 0.2       , 0.2       , 0.2       ],
        [0.        , 0.25      , 0.25      , 0.25      , 0.25      ],
        [0.        , 0.        , 0.33333333, 0.33333333, 0.33333333],
        [0.        , 0.        , 0.        , 0.5       , 0.5       ],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

def test_left_right_fit_updates_random_transitions():
    """Check fitting with a random transition matrix of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.transitions
    assert_equal(hmm.transitions, np.array([
        [0.19252534, 0.15767581, 0.47989976, 0.01708551, 0.15281357],
        [0.        , 0.21269278, 0.26671807, 0.16241481, 0.35817434],
        [0.        , 0.        , 0.33753566, 0.19947995, 0.46298439],
        [0.        , 0.        , 0.        , 0.39158159, 0.60841841],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

def test_ergodic_fit_updates_uniform_initial():
    """Check fitting with a uniform initial state distribution of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    before = hmm.initial
    hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_ergodic_fit_updates_random_initial():
    """Check fitting with a random initial state distribution of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))
    before = hmm.initial
    hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_ergodic_fit_updates_uniform_transitions():
    """Check fitting with a uniform transition matrix of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))
    before = hmm.transitions
    hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

def test_ergodic_fit_updates_random_transitions():
    """Check fitting with a random transition matrix of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.19252534, 0.15767581, 0.47989976, 0.01708551, 0.15281357],
        [0.19375092, 0.16425506, 0.21828034, 0.11397708, 0.30973661],
        [0.46906977, 0.02941216, 0.17137502, 0.0333193 , 0.29682374],
        [0.21312406, 0.35221103, 0.08556524, 0.06613143, 0.28296824],
        [0.05212313, 0.3345513 , 0.17192948, 0.16379392, 0.27760217]]))
    before = hmm.transitions
    hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

def test_linear_fit_updates_uniform_initial():
    """Check fitting with a uniform initial state distribution of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    before = hmm.initial
    hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_linear_fit_updates_random_initial():
    """Check fitting with a random initial state distribution of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))
    before = hmm.initial
    hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_linear_fit_updates_uniform_transitions():
    """Check fitting with a uniform transition matrix of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.transitions
    assert_equal(hmm.transitions, np.array([
        [0.5, 0.5, 0. , 0. , 0. ],
        [0. , 0.5, 0.5, 0. , 0. ],
        [0. , 0. , 0.5, 0.5, 0. ],
        [0. , 0. , 0. , 0.5, 0.5],
        [0. , 0. , 0. , 0. , 1. ]
    ]))
    hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

def test_linear_fit_updates_random_transitions():
    """Check fitting with a random transition matrix of a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    before = hmm.transitions
    assert_equal(hmm.transitions, np.array([
        [0.54975645, 0.45024355, 0.        , 0.        , 0.        ],
        [0.        , 0.96562169, 0.03437831, 0.        , 0.        ],
        [0.        , 0.        , 0.29607315, 0.70392685, 0.        ],
        [0.        , 0.        , 0.        , 0.42938524, 0.57061476],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

# ================ #
# GMMHMM.forward() #
# ================ #

def test_forward_without_fit():
    """Forward algorithm without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.forward(x)
    assert str(e.value) == 'The model must be fitted before running the forward algorithm'

def test_left_right_forward():
    """Forward algorithm on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert isinstance(hmm.forward(x), float)

def test_ergodic_forward():
    """Forward algorithm on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert isinstance(hmm.forward(x), float)

def test_linear_forward():
    """Forward algorithm on a linear HMM"""
    hmm = deepcopy(hmm_lin)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert isinstance(hmm.forward(x), float)

# ======================= #
# GMMHMM.label (property) #
# ======================= #

def test_label():
    assert deepcopy(hmm_lr).label == 'c1'

# ========================== #
# GMMHMM.n_states (property) #
# ========================== #

def test_n_states():
    assert deepcopy(hmm_lr).n_states == 5

# ============================== #
# GMMHMM.n_components (property) #
# ============================== #

def test_n_components():
    assert deepcopy(hmm_lr).n_components == 1

# ================================= #
# GMMHMM.covariance_type (property) #
# ================================= #

def test_covariance_type():
    assert deepcopy(hmm_lr).covariance_type == 'full'

# ======================== #
# GMMHMM.n_seqs (property) #
# ======================== #

def test_n_seqs_without_fit():
    """Number of sequences without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.n_seqs
    assert str(e.value) == 'The model has not been fitted and has not seen any observation sequences'

def test_n_seqs_with_fit():
    """Number of sequences after fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    hmm.fit(X)
    assert hmm.n_seqs == 3

# ========================= #
# GMMHMM.initial (property) #
# ========================= #

def test_initial_without_setting():
    """Get initial state distribution without setting it"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.initial
    assert str(e.value) == 'No initial state distribution has been defined'

# ============================= #
# GMMHMM.transitions (property) #
# ============================= #

def test_transitions_without_setting():
    """Get transition matrix without setting it"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.transitions
    assert str(e.value) == 'No transition matrix has been defined'

# ======================= #
# GMMHMM.initial (setter) #
# ======================= #

def test_left_right_initial_left_right():
    """Set an initial state distribution generated by a left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_left_right_initial_ergodic():
    """Set an initial state distribution generated by a left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_left_right_initial_linear():
    """Set an initial state distribution generated by a left-right topology on an linear HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LinearTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_ergodic_initial_left_right():
    """Set an initial state distribution generated by an ergodic topology on a left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_ergodic_initial_ergodic():
    """Set an initial state distribution generated by an ergodic topology on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_ergodic_initial_linear():
    """Set an initial state distribution generated by an ergodic topology on a linear HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LinearTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_linear_initial_left_right():
    """Set an initial state distribution generated by a linear topology on a left-right HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_linear_initial_ergodic():
    """Set an initial state distribution generated by a linear topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_linear_initial_linear():
    """Set an initial state distribution generated by a linear topology on an linear HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LinearTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

# =========================== #
# GMMHMM.transitions (setter) #
# =========================== #

def test_left_right_transitions_left_right():
    """Set a transition matrix generated by a left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)

def test_left_right_transitions_ergodic():
    """Set a transition matrix generated by a left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions = transitions
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_left_right_transitions_linear():
    """Set a transition matrix generated by a left-right topology on a linear HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)

def test_ergodic_transitions_left_right():
    """Set a transition matrix generated by an ergodic topology on a left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.warns(UserWarning) as w:
        hmm.transitions = transitions
    assert w[0].message.args[0] == 'Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned'
    assert_equal(hmm.transitions, transitions)

def test_ergodic_transitions_ergodic():
    """Set a transition matrix generated by an ergodic topology on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)

def test_ergodic_transitions_linear():
    """Set a transition matrix generated by an ergodic topology on a linear HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.warns(UserWarning) as w:
        hmm.transitions = transitions
    assert w[0].message.args[0] == 'Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned'
    assert_equal(hmm.transitions, transitions)

def test_linear_transitions_left_right():
    """Set a transition matrix generated by a linear topology on a left-right HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions = transitions
    assert str(e.value) == 'Linear transition matrix must only consist of a diagonal and upper diagonal'

def test_linear_transitions_ergodic():
    """Set a transition matrix generated by a linear topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions = transitions
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_linear_transitions_linear():
    """Set a transition matrix generated by a linear topology on a linear HMM"""
    hmm = deepcopy(hmm_lin)
    topology = _LinearTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)


import pytest, warnings, os, numpy as np, pickle
from copy import deepcopy
from multiprocessing import cpu_count
from sequentia.classifiers import KNNClassifier
from ....support import assert_equal, assert_all_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

classes = ['c0', 'c1', 'c2', 'c3', 'c4']

# Create some sample data
X = [rng.random((10 * i, 3)) for i in range(1, 7)]
y = ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']
x = X[0]

clfs = {
    'k=1': KNNClassifier(k=1, classes=classes, random_state=rng),
    'k=2': KNNClassifier(k=2, classes=classes, random_state=rng),
    'k=3': KNNClassifier(k=3, classes=classes, random_state=rng),
    'weighted': KNNClassifier(k=3, classes=classes, weighting=(lambda x: np.exp(-x)), random_state=rng),
    'independent': KNNClassifier(k=1, classes=classes, independent=True, random_state=rng)
}

for _, clf in clfs.items():
    clf.fit(X, y)

# =================== #
# KNNClassifier.fit() #
# =================== #

def test_fit_sets_attributes():
    """Check that fitting sets the hidden attributes"""
    clf = clfs['k=1']
    clf.fit(X, y)
    assert_all_equal(clf._X, X)
    assert_equal(clf._y, clf._encoder.transform(y))
    assert clf._n_features == 3

# ======================= #
# KNNClassifier.predict() #
# ======================= #

def test_predict_without_fit():
    """Predict without fitting the model"""
    with pytest.raises(RuntimeError) as e:
        KNNClassifier(k=1, classes=classes).predict(x, verbose=False)
    assert str(e.value) == 'The classifier needs to be fitted before predictions are made'

def test_predict_single_k1_verbose(capsys):
    """Verbosely predict a single observation sequence (k=1)"""
    prediction = clfs['k=1'].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k1_no_verbose(capsys):
    """Silently predict a single observation sequence (k=1)"""
    prediction = clfs['k=1'].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k2_verbose(capsys):
    """Verbosely predict a single observation sequence (k=2)"""
    prediction = clfs['k=2'].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k2_no_verbose(capsys):
    """Silently predict a single observation sequence (k=2)"""
    prediction = clfs['k=2'].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k3_verbose(capsys):
    """Verbosely predict a single observation sequence (k=3)"""
    prediction = clfs['k=3'].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k3_no_verbose(capsys):
    """Silently predict a single observation sequence (k=3)"""
    prediction = clfs['k=3'].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_weighted_verbose(capsys):
    """Verbosely predict a single observation sequence (weighted)"""
    prediction = clfs['weighted'].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_weighted_no_verbose(capsys):
    """Silently predict a single observation sequence (weighted)"""
    prediction = clfs['weighted'].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_independent_verbose(capsys):
    """Verbosely predict a single observation sequence with independent warping"""
    prediction = clfs['independent'].predict(x, verbose=True)
    assert 'Calculating distances' in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_single_k1_no_verbose(capsys):
    """Silently predict a single observation sequence with independent warping"""
    prediction = clfs['independent'].predict(x, verbose=False)
    assert 'Calculating distances' not in capsys.readouterr().err
    assert prediction == 'c1'

def test_predict_multiple_k1_verbose(capsys):
    """Verbosely predict multiple observation sequences (k=1)"""
    predictions = clfs['k=1'].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_multiple_k1_no_verbose(capsys):
    """Silently predict multiple observation sequences (k=1)"""
    predictions = clfs['k=1'].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_multiple_k2_verbose(capsys):
    """Verbosely predict multiple observation sequences (k=2)"""
    predictions = clfs['k=2'].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert len(predictions) == 6

def test_predict_multiple_k2_no_verbose(capsys):
    """Silently predict multiple observation sequences (k=2)"""
    predictions = clfs['k=2'].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert len(predictions) == 6

def test_predict_multiple_k3_verbose(capsys):
    """Verbosely predict multiple observation sequences (k=3)"""
    predictions = clfs['k=3'].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c1', 'c1', 'c0', 'c1']

def test_predict_multiple_k3_no_verbose(capsys):
    """Silently predict multiple observation sequences (k=3)"""
    predictions = clfs['k=3'].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c1', 'c1', 'c0', 'c1']

def test_predict_multiple_weighted_verbose(capsys):
    """Verbosely predict multiple observation sequences (weighted)"""
    predictions = clfs['weighted'].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c0', 'c1']

def test_predict_multiple_weighted_no_verbose(capsys):
    """Silently predict multiple observation sequences (weighted)"""
    predictions = clfs['weighted'].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c0', 'c1']

def test_predict_multiple_independent_verbose(capsys):
    """Verbosely predict multiple observation sequences with independent warping"""
    predictions = clfs['independent'].predict(X, verbose=True)
    assert 'Classifying examples' in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_multiple_independent_no_verbose(capsys):
    """Silently predict multiple observation sequences with independent warping"""
    predictions = clfs['independent'].predict(X, verbose=False)
    assert 'Classifying examples' not in capsys.readouterr().err
    assert list(predictions) == ['c1', 'c1', 'c0', 'c1', 'c1', 'c0']

def test_predict_single():
    """Predict a single observation sequence and don't return the original labels"""
    prediction = clfs['k=3'].predict(x, verbose=False, original_labels=False)
    assert prediction == 1

def test_predict_single_original_labels():
    """Predict a single observation sequence and return the original labels"""
    prediction = clfs['k=3'].predict(x, verbose=False, original_labels=True)
    assert prediction == 'c1'

def test_predict_multiple():
    """Predict multiple observation sequences and don't return the original labels"""
    predictions = clfs['k=3'].predict(X, verbose=False, original_labels=False)
    assert list(predictions) == [1, 1, 1, 1, 0, 1]

def test_predict_multiple_original_labels():
    """Predict multiple observation sequences and return the original labels"""
    predictions = clfs['k=3'].predict(X, verbose=False, original_labels=True)
    assert list(predictions) == ['c1', 'c1', 'c1', 'c1', 'c0', 'c1']

# ======================== #
# KNNClassifier.evaluate() #
# ======================== #

def test_evaluate():
    """Evaluate performance on some observation sequences and labels"""
    acc, cm = clfs['k=3'].evaluate(X, y)
    assert acc == 0.5
    assert_equal(cm, np.array([
        [0, 2, 0, 0, 0],
        [1, 3, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0]
    ]))

# ==================== #
# KNNClassifier.save() #
# ==================== #

def test_save_unfitted():
    """Save an unfitted KNNClassifier object."""
    try:
        with pytest.raises(RuntimeError) as e:
            KNNClassifier(k=1, classes=classes).save('test.pkl')
        assert str(e.value) == 'The classifier needs to be fitted before it can be saved'
    finally:
        if os.path.exists('test.pkl'):
            os.remove('test.pkl')

def test_save_fitted():
    """Save a fitted KNNClassifier object."""
    try:
        clfs['weighted'].save('test.pkl')
        assert os.path.isfile('test.pkl')
    finally:
        os.remove('test.pkl')

# ==================== #
# KNNClassifier.load() #
# ==================== #

def test_load_invalid_format():
    """Load a KNNClassifier from an illegally formatted file"""
    try:
        with open('test', 'w') as f:
            f.write('illegal')
        with pytest.raises(pickle.UnpicklingError) as e:
            KNNClassifier.load('test')
    finally:
        os.remove('test')

def test_load_invalid_object_type():
    """Load a KNNClassifier from a pickle file with invalid object type"""
    try:
        with open('test.pkl', 'wb') as file:
            pickle.dump(0, file)
        with pytest.raises(TypeError) as e:
            KNNClassifier.load('test.pkl')
        assert str(e.value) == 'Expected deserialized object to be a dictionary - make sure the object was serialized with the save() function'
    finally:
        os.remove('test.pkl')

def test_load_missing_keys():
    """Load a KNNClassifier from a pickled dict with invalid keys"""
    try:
        with open('test.pkl', 'wb') as file:
            pickle.dump({'test': 0}, file)
        with pytest.raises(ValueError) as e:
            KNNClassifier.load('test.pkl')
        assert str(e.value) == 'Missing keys in deserialized object dictionary – make sure the object was serialized with the save() function'
    finally:
        os.remove('test.pkl')

def test_load_valid_no_weighting():
    """Load a serialized KNNClassifier with the default weighting function"""
    try:
        clfs['k=3'].save('test.pkl')
        clf = KNNClassifier.load('test.pkl')
        # Check that all fields are still the same
        assert isinstance(clf, KNNClassifier)
        assert clf._k == 3
        assert list(clf._encoder.classes_) == classes
        assert clf._window == 1.
        assert clf._use_c == False
        assert clf._independent == False
        assert deepcopy(clf._random_state).normal() == deepcopy(rng).normal()
        assert_all_equal(clf._X, X)
        assert_equal(clf._y, clf._encoder.transform(y))
        assert clf._n_features == 3
        # Check that weighting functions are the same for x=0 to x=1000
        xs = np.arange(1000, step=0.1)
        weighting = lambda x: np.ones(x.size)
        assert_equal(clf._weighting(xs), weighting(xs))
    finally:
        os.remove('test.pkl')

def test_load_valid_weighting():
    """Load a serialized KNNClassifier with a custom weighting function"""
    try:
        clfs['weighted'].save('test.pkl')
        clf = KNNClassifier.load('test.pkl')
        # Check that all fields are still the same
        assert isinstance(clf, KNNClassifier)
        assert clf._k == 3
        assert list(clf._encoder.classes_) == classes
        assert clf._window == 1.
        assert clf._use_c == False
        assert clf._independent == False
        assert deepcopy(clf._random_state).normal() == deepcopy(rng).normal()
        assert_all_equal(clf._X, X)
        assert_equal(clf._y, clf._encoder.transform(y))
        assert clf._n_features == 3
        # Check that weighting functions are the same for x=0 to x=1000
        xs = np.arange(1000, step=0.1)
        weighting = lambda x: np.exp(-x)
        assert_equal(clf._weighting(xs), weighting(xs))
    finally:
        os.remove('test.pkl')


import numpy as np
from copy import copy
from tqdm.auto import tqdm
from .transforms import Transform
from ..internals import _Validator

__all__ = ['Preprocess']

class Preprocess:
    """A pipeline of preprocessing transformations.

    Parameters
    ----------
    steps: array-like of Transform
        An ordered collection of preprocessing transformations.

    Examples
    --------
    >>> # Create some sample data
    >>> X = [np.random.random((20 * i, 3)) for i in range(1, 4)]
    >>> # Create the Preprocess object
    >>> pre = Preprocess([
    >>>     TrimZeros(),
    >>>     Center(),
    >>>     Standardize(),
    >>>     Filter(window_size=5, method='median'),
    >>>     Downsample(factor=5, method='decimate')
    >>> ])
    >>> # View a summary of the preprocessing steps
    >>> pre.summary()
    >>> # Transform the data applying transformations in order
    >>> X = pre(X)
    """

    def __init__(self, steps):
        self._val = _Validator()
        steps = list(self._val.iterable(steps, 'transformation steps'))
        if not all(isinstance(transform, Transform) for transform in steps):
            raise TypeError('Expected all transformation steps to be Transform objects')
        self.steps = steps

    def transform(self, X, verbose=False):
        """Applies the preprocessing transformations to the provided input observation sequence(s).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether or not to display a progress bar when applying transformations.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float) or list of :class:`numpy:numpy.ndarray` (float)
            The input observation sequence(s) with preprocessing transformations applied in order.
        """
        X_t = copy(X)
        pbar = tqdm(self.steps, desc='Applying transformations', disable=not(verbose and len(self.steps) > 1), leave=True, ncols='100%')
        for step in pbar:
            pbar.set_description("Applying transformations - {}".format(step._describe()))
            X_t = step.transform(X_t, verbose=False)
        return X_t

    def __call__(self, X, verbose=False):
        """Alias of the :meth:`transform` method.

        See Also
        --------
        transform: Applies the transformation.
        """
        return self.transform(X, verbose)

    def _fit(self, X, verbose):
        """Fit the preprocessing transformations with the provided observation sequence(s).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether or not to display a progress bar when fitting transformations.
        """
        X = self._val.observation_sequences(X, allow_single=True)
        X_t = copy(X)
        pbar = tqdm(self.steps, desc='Fitting transformations', disable=not(verbose and len(self.steps) > 1), leave=True, ncols='100%')
        for step in pbar:
            pbar.set_description("Fitting transformations - {}".format(step._describe()))
            X_t = step.fit_transform(X_t, verbose=False)
        return X_t

    def fit(self, X, verbose=False):
        """Fit the preprocessing transformations with the provided observation sequence(s).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether or not to display a progress bar when fitting transformations.
        """
        self._fit(X, verbose)

    def fit_transform(self, X, verbose=False):
        """Fit the preprocessing transformations with the provided observation sequence(s) and transform them.

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether or not to display a progress bar when fitting and applying transformations.

        Returns
        -------
        transformed: :class:`numpy:numpy.ndarray` (float) or list of :class:`numpy:numpy.ndarray` (float)
            The input observation sequence(s) with preprocessing transformations applied in order.
        """
        return self._fit(X, verbose)

    def summary(self):
        """Displays an ordered summary of the preprocessing transformations."""
        if len(self.steps) == 0:
            raise RuntimeError('At least one preprocessing transformation is required')

        steps = []

        for i, step in enumerate(self.steps, start=1):
            class_name, description = step.__class__.__name__, step._describe()
            steps.append(('{}. {}'.format(i, class_name), '   {}'.format(description)))

        title = 'Preprocessing summary:'
        length = max(max(len(h), 0 if b is None else len(b)) for h, b in steps)
        length = len(title) if length < len(title) else length

        print(title.center(length, ' '))
        print('=' * length)
        for i, (head, body) in enumerate(steps):
            print(head)
            if body is not None:
                print(body)
            if i != len(steps) - 1:
                print('-' * length)
        print('=' * length)


import pytest, numpy as np
from sequentia.preprocessing import *
from ...support import assert_equal, assert_all_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# Sample data
X = rng.random((7, 2))
Xs = [i * rng.random((3 * i, 2)) for i in range(1, 4)]

# Length equalizing preprocessor
equalize = Preprocess([Equalize()])

# Zero-trimming preprocessor
trim = Preprocess([TrimZeros()])

# Min-max scaling preprocessor
min_max_scale_kwargs = {'scale': (-5, 5), 'independent': False}
min_max_scale = Preprocess([MinMaxScale(**min_max_scale_kwargs)])

# Centering preprocessor
cent = Preprocess([Center()])

# Standardizing preprocessor
standard = Preprocess([Standardize()])

# Downsampling preprocessor
down_kwargs = {'factor': 3, 'method': 'decimate'}
down = Preprocess([Downsample(**down_kwargs)])

# Filtering preprocessor
filt_kwargs = {'window_size': 3, 'method': 'median'}
filt = Preprocess([Filter(**filt_kwargs)])

# Combined preprocessor
combined = Preprocess([
    Equalize(),
    TrimZeros(),
    MinMaxScale(**min_max_scale_kwargs),
    Center(),
    Standardize(),
    Filter(**filt_kwargs),
    Downsample(**down_kwargs)
])

# ======== #
# Equalize #
# ======== #

def test_equalize_single():
    """Applying length equalizing to a single observation sequence"""
    assert_equal(equalize(X), Equalize()(X))

def test_equalize_multiple():
    """Applying length equalizing to multiple observation sequences"""
    assert_all_equal(equalize(Xs), Equalize()(Xs))

def test_equalize_summary(capsys):
    """Summary of a length equalizing transformation"""
    equalize.summary()
    assert capsys.readouterr().out == (
        '   Preprocessing summary:   \n'
        '============================\n'
        '1. Equalize\n'
        '   Equalize sequence lengths\n'
        '============================\n'
    )

# ========= #
# TrimZeros #
# ========= #

def test_trim_zeros_single():
    """Applying zero-trimming to a single observation sequence"""
    assert_equal(trim(X), TrimZeros()(X))

def test_trim_zeros_multiple():
    """Applying zero-trimming to multiple observation sequences"""
    assert_all_equal(trim(Xs), TrimZeros()(Xs))

def test_trim_zeros_summary(capsys):
    """Summary of a zero-trimming transformation"""
    trim.summary()
    assert capsys.readouterr().out == (
        '   Preprocessing summary:  \n'
        '===========================\n'
        '1. TrimZeros\n   '
        'Remove zero-observations\n'
        '===========================\n'
    )

# =========== #
# MinMaxScale #
# =========== #

def test_min_max_scale_single():
    """Applying min-max scaling to a single observation sequence"""
    assert_equal(min_max_scale(X), MinMaxScale(**min_max_scale_kwargs)(X))

def test_min_max_scale_multiple():
    """Applying min-max scaling to multiple observation sequences"""
    assert_all_equal(min_max_scale(Xs), MinMaxScale(**min_max_scale_kwargs)(Xs))

def test_min_max_scale_summary(capsys):
    """Summary of a min-max scaling transformation"""
    min_max_scale.summary()
    assert capsys.readouterr().out == (
        '        Preprocessing summary:       \n'
        '=====================================\n'
        '1. MinMaxScale\n'
        '   Min-max scaling into range (-5, 5)\n'
        '=====================================\n'
    )

# ====== #
# Center #
# ====== #

def test_center_single():
    """Applying centering to a single observation sequence"""
    assert_equal(cent(X), Center()(X))

def test_center_multiple():
    """Applying centering to multiple observation sequences"""
    assert_all_equal(cent(Xs), Center()(Xs))

def test_center_summary(capsys):
    """Summary of a centering transformation"""
    cent.summary()
    assert capsys.readouterr().out == (
        '              Preprocessing summary:              \n'
        '==================================================\n'
        '1. Center\n'
        '   Centering around mean (zero mean) (independent)\n'
        '==================================================\n'
    )

# =========== #
# Standardize #
# =========== #

def test_standardize_single():
    """Applying standardization to a single observation sequence"""
    assert_equal(standard(X), Standardize()(X))

def test_standardize_multiple():
    """Applying standardization to multiple observation sequences"""
    assert_all_equal(standard(Xs), Standardize()(Xs))

def test_standardize_summary(capsys):
    """Summary of a standardizing transformation"""
    standard.summary()
    assert capsys.readouterr().out == (
        '                   Preprocessing summary:                   \n'
        '============================================================\n'
        '1. Standardize\n'
        '   Standard scaling (zero mean, unit variance) (independent)\n'
        '============================================================\n'
    )

# ========== #
# Downsample #
# ========== #

def test_downsample_single():
    """Applying downsampling to a single observation sequence"""
    assert_equal(down(X), Downsample(**down_kwargs)(X))

def test_downsample_multiple():
    """Applying downsampling to multiple observation sequences"""
    assert_all_equal(down(Xs), Downsample(**down_kwargs)(Xs))

def test_downsample_summary(capsys):
    """Summary of a downsampling transformation"""
    down.summary()
    assert capsys.readouterr().out == (
        '         Preprocessing summary:         \n'
        '========================================\n'
        '1. Downsample\n'
        '   Decimation downsampling with factor 3\n'
        '========================================\n'
    )

# ====== #
# Filter #
# ====== #

def test_filter_single():
    """Applying filtering to a single observation sequence"""
    assert_equal(filt(X), Filter(**filt_kwargs)(X))

def test_filter_multiple():
    """Applying filtering to multiple observation sequences"""
    assert_all_equal(filt(Xs), Filter(**filt_kwargs)(Xs))

def test_filter_summary(capsys):
    """Summary of a filtering transformation"""
    filt.summary()
    assert capsys.readouterr().out == (
        '        Preprocessing summary:        \n'
        '======================================\n'
        '1. Filter\n'
        '   Median filtering with window-size 3\n'
        '======================================\n'
    )

# ======================== #
# Combined transformations #
# ======================== #

combined = Preprocess([
    Equalize(),
    TrimZeros(),
    MinMaxScale(**min_max_scale_kwargs),
    Center(),
    Standardize(),
    Filter(**filt_kwargs),
    Downsample(**down_kwargs)
])

def test_combined_single():
    """Applying combined transformations to a single observation sequence"""
    X_pre = X
    X_pre = Equalize()(X_pre)
    X_pre = TrimZeros()(X_pre)
    X_pre = MinMaxScale(**min_max_scale_kwargs)(X_pre)
    X_pre = Center()(X_pre)
    X_pre = Standardize()(X_pre)
    X_pre = Filter(**filt_kwargs)(X_pre)
    X_pre = Downsample(**down_kwargs)(X_pre)
    assert_equal(combined(X), X_pre)

def test_combined_multiple():
    """Applying combined transformations to multiple observation sequences"""
    Xs_pre = Xs
    Xs_pre = Equalize()(Xs_pre)
    Xs_pre = TrimZeros()(Xs_pre)
    Xs_pre = MinMaxScale(**min_max_scale_kwargs)(Xs_pre)
    Xs_pre = Center()(Xs_pre)
    Xs_pre = Standardize()(Xs_pre)
    Xs_pre = Filter(**filt_kwargs)(Xs_pre)
    Xs_pre = Downsample(**down_kwargs)(Xs_pre)
    assert_all_equal(combined(Xs), Xs_pre)

def test_combined_summary(capsys):
    """Summary with combined transformations applied"""
    combined.summary()
    assert capsys.readouterr().out == (
        '                   Preprocessing summary:                   \n'
        '============================================================\n'
        '1. Equalize\n'
        '   Equalize sequence lengths\n'
        '------------------------------------------------------------\n'
        '2. TrimZeros\n'
        '   Remove zero-observations\n'
        '------------------------------------------------------------\n'
        '3. MinMaxScale\n'
        '   Min-max scaling into range (-5, 5)\n'
        '------------------------------------------------------------\n'
        '4. Center\n'
        '   Centering around mean (zero mean) (independent)\n'
        '------------------------------------------------------------\n'
        '5. Standardize\n'
        '   Standard scaling (zero mean, unit variance) (independent)\n'
        '------------------------------------------------------------\n'
        '6. Filter\n'
        '   Median filtering with window-size 3\n'
        '------------------------------------------------------------\n'
        '7. Downsample\n'
        '   Decimation downsampling with factor 3\n'
        '============================================================\n'
    )

def test_empty_summary():
    """Summary without any transformations applied"""
    with pytest.raises(RuntimeError) as e:
        Preprocess([]).summary()
    assert str(e.value) == 'At least one preprocessing transformation is required'


import tqdm, tqdm.auto, numpy as np, pickle
from joblib import Parallel, delayed
from multiprocessing import cpu_count
from .gmmhmm import GMMHMM
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import LabelEncoder
from ...internals import _Validator

class HMMClassifier:
    """A classifier that combines individual :class:`~GMMHMM` objects, which each model isolated sequences from a different class.

    Attributes
    ----------
    models: list of GMMHMM
        A collection of the :class:`~GMMHMM` objects to use for classification.

    encoder: sklearn.preprocessing.LabelEncoder
        The label encoder fitted on the set of ``classes`` provided during instantiation.

    classes: numpy.ndarray (str/numeric)
        The complete set of possible classes/labels.
    """

    def __init__(self):
        self._val = _Validator()

    def fit(self, models):
        """Fits the classifier with a collection of :class:`~GMMHMM` objects.

        Parameters
        ----------
        models: array-like of GMMHMM
            A collection of :class:`~GMMHMM` objects to use for classification.
        """

        models = list(self._val.iterable(models, 'models'))
        if not all(isinstance(model, GMMHMM) for model in models):
            raise TypeError('Expected all models to be GMMHMM objects')

        if len(models) > 0:
            self._models = models
        else:
            raise RuntimeError('The classifier must be fitted with at least one HMM')

        self._encoder = LabelEncoder()
        self._encoder.fit([model.label for model in models])

    def predict(self, X, prior='frequency', return_scores=False, original_labels=True, verbose=True, n_jobs=1):
        """Predicts the label for an observation sequence (or multiple sequences) according to maximum likelihood or posterior scores.

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        prior: {'frequency', 'uniform'} or array-like of float
            How the prior probability for each model is calculated to perform MAP estimation by scoring with
            the joint probability (or un-normalized posterior) :math:`\\mathbb{P}(O, \\lambda_c)=\\mathbb{P}(O|\\lambda_c)\\mathbb{P}(\\lambda_c)`.

            - `'frequency'`: Calculate the prior probability :math:`\\mathbb{P}(\\lambda_c)` as the proportion of training examples in class :math:`c`.
            - `'uniform'`: Set the priors uniformly such that :math:`\\mathbb{P}(\\lambda_c)=\\frac{1}{C}` for each class :math:`c\\in\\{1,\\ldots,C\\}` (**equivalent to ignoring the prior**).

            Alternatively, class prior probabilities can be specified in an iterable of floats, e.g. `[0.1, 0.3, 0.6]`.

        return_scores: bool
            Whether to return the scores of each model on the observation sequence(s).

        original_labels: bool
            Whether to inverse-transform the labels to their original encoding.

        verbose: bool
            Whether to display a progress bar or not.

            .. note::
                If both ``verbose=True`` and ``n_jobs > 1``, then the progress bars for each process
                are always displayed in the console, regardless of where you are running this function from
                (e.g. a Jupyter notebook).

        n_jobs: int > 0 or -1
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        prediction(s): str/numeric or :class:`numpy:numpy.ndarray` (str/numeric)
            The predicted label(s) for the observation sequence(s).

            If ``original_labels`` is true, then the returned labels are inverse-transformed into their original encoding.

        scores: :class:`numpy:numpy.ndarray` (float)
            An :math:`N\\times M` matrix of scores (log un-normalized posteriors), for each of the :math:`N` observation sequences,
            for each of the :math:`M` HMMs. Only returned if ``return_scores`` is true.
        """
        try:
            self._models
        except AttributeError as e:
            raise AttributeError('The classifier needs to be fitted before predictions are made') from e

        X = self._val.observation_sequences(X, allow_single=True)
        if not isinstance(prior, str):
            self._val.iterable(prior, 'prior')
            assert len(prior) == len(self._models), 'There must be a class prior for each HMM or class'
            assert all(isinstance(p, (int, float)) for p in prior), 'Class priors must be numerical'
            assert all(0. <= p <= 1. for p in prior), 'Class priors must each be between zero and one'
            assert np.isclose(sum(prior), 1.), 'Class priors must form a probability distribution by summing to one'
        else:
            self._val.one_of(prior, ['frequency', 'uniform'], desc='prior')
        self._val.boolean(return_scores, desc='return_scores')
        self._val.boolean(original_labels, desc='original_labels')
        self._val.boolean(verbose, desc='verbose')
        self._val.restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        # Create look-up for prior probabilities
        if prior == 'frequency':
            total_seqs = sum(model.n_seqs for model in self._models)
            prior = {model.label:(model.n_seqs / total_seqs) for model in self._models}
        elif prior == 'uniform':
            prior = {model.label:(1. / len(self._models)) for model in self._models}
        else:
            prior = {model.label:prior[self._encoder.transform([model.label]).item()] for model in self._models}

        # Convert single observation sequence to a singleton list
        X = [X] if isinstance(X, np.ndarray) else X

        # Lambda for calculating the log un-normalized posteriors as a sum of the log forward probabilities (likelihoods) and log priors
        posteriors = lambda x: np.array([model.forward(x) + np.log(prior[model.label]) for model in self._models])

        # Calculate log un-normalized posteriors as a sum of the log forward probabilities (likelihoods) and log priors
        # Perform the MAP classification rule and return labels to original encoding if necessary
        n_jobs = min(cpu_count() if n_jobs == -1 else n_jobs, len(X))
        X_chunks = [list(chunk) for chunk in np.array_split(np.array(X, dtype=object), n_jobs)]
        scores = Parallel(n_jobs=n_jobs)(delayed(self._chunk_predict)(i+1, posteriors, chunk, verbose) for i, chunk in enumerate(X_chunks))
        scores = np.concatenate(scores)
        best_idxs = np.atleast_1d(scores.argmax(axis=1))
        labels = self._encoder.inverse_transform(best_idxs) if original_labels else best_idxs

        if len(X) == 1:
            return (labels.item(), scores.flatten()) if return_scores else labels.item()
        else:
            return (labels, scores) if return_scores else labels

    def evaluate(self, X, y, prior='frequency', verbose=True, n_jobs=1):
        """Evaluates the performance of the classifier on a batch of observation sequences and their labels.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            A list of multiple observation sequences.

        y: array-like of str/numeric
            An iterable of labels for the observation sequences.

        prior: {'frequency', 'uniform'} or array-like of float
            How the prior probability for each model is calculated to perform MAP estimation by scoring with
            the joint probability (or un-normalized posterior) :math:`\\mathbb{P}(O, \\lambda_c)=\\mathbb{P}(O|\\lambda_c)\\mathbb{P}(\\lambda_c)`.

            - `'frequency'`: Calculate the prior probability :math:`\\mathbb{P}(\\lambda_c)` as the proportion of training examples in class :math:`c`.
            - `'uniform'`: Set the priors uniformly such that :math:`\\mathbb{P}(\\lambda_c)=\\frac{1}{C}` for each class :math:`c\\in\\{1,\\ldots,C\\}` (**equivalent to ignoring the prior**).

            Alternatively, class prior probabilities can be specified in an iterable of floats, e.g. `[0.1, 0.3, 0.6]`.

        verbose: bool
            Whether to display a progress bar or not.

            .. note::
                If both ``verbose=True`` and ``n_jobs > 1``, then the progress bars for each process
                are always displayed in the console, regardless of where you are running this function from
                (e.g. a Jupyter notebook).

        n_jobs: int > 0 or -1
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        accuracy: float
            The categorical accuracy of the classifier on the observation sequences.

        confusion: :class:`numpy:numpy.ndarray` (int)
            The confusion matrix representing the discrepancy between predicted and actual labels.
        """
        X, y = self._val.observation_sequences_and_labels(X, y)
        predictions = self.predict(X, prior=prior, return_scores=False, original_labels=False, verbose=verbose, n_jobs=n_jobs)
        cm = confusion_matrix(self._encoder.transform(y), predictions, labels=self._encoder.transform(self._encoder.classes_))
        return np.sum(np.diag(cm)) / np.sum(cm), cm

    def save(self, path):
        """Serializes the :class:`HMMClassifier` object by pickling.

        Parameters
        ----------
        path: str
            File path (usually with `.pkl` extension) to store the serialized :class:`HMMClassifier` object.
        """
        try:
            self._models
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted before it can be saved')

        with open(path, 'wb') as file:
            pickle.dump(self, file)

    @classmethod
    def load(cls, path):
        """Deserializes a :class:`HMMClassifier` object which was serialized with the :meth:`save` function.

        Parameters
        ----------
        path: str
            File path of the serialized data generated by the :meth:`save` method.

        Returns
        -------
        deserialized: :class:`HMMClassifier`
            The deserialized HMM classifier object.
        """
        with open(path, 'rb') as file:
            return pickle.load(file)

    def _chunk_predict(self, process, posteriors, chunk, verbose): # Requires fit
        """Makes predictions (scores) for a chunk of the observation sequences, for a given subprocess."""
        return np.array([posteriors(x) for x in tqdm.auto.tqdm(
            chunk, desc='Classifying examples (process {})'.format(process),
            disable=not(verbose), position=process-1
        )])

    @property
    def models(self):
        try:
            return self._models
        except AttributeError as e:
            raise AttributeError('No models available - the classifier must be fitted first') from e

    @property
    def encoder(self):
        try:
            return self._encoder
        except AttributeError as e:
            raise AttributeError('No label encoder has been defined - the classifier must be fitted first') from e

    @property
    def classes(self):
        return self.encoder.classes_

    def __repr__(self):
        module = self.__class__.__module__
        out = '{}{}('.format('' if module == '__main__' else '{}.'.format(module), self.__class__.__name__)
        try:
            self._models
            out += 'models=[\n  '
            out += ',\n  '.join(repr(model) for model in self._models)
            out += '\n]'
        except AttributeError:
            pass
        return out + ')'


import numpy as np, hmmlearn.hmm
from .topologies.ergodic import _ErgodicTopology
from .topologies.left_right import _LeftRightTopology
from .topologies.linear import _LinearTopology
from ...internals import _Validator

class GMMHMM:
    """A hidden Markov model (with Gaussian Mixture Model emissions)
    representing a single isolated sequence class.

    Parameters
    ----------
    label: str or numeric
        A label for the model, corresponding to the class being represented.

    n_states: int > 0
        The number of states for the model.

    n_components: int > 0
        The number of mixture components used in the emission distribution for each state.

    covariance_type: {'spherical', 'diag', 'full', 'tied'}
        The covariance matrix type for emission distributions.

    topology: {'ergodic', 'left-right', 'linear'}
        The topology for the model.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Attributes
    ----------
    label: str or numeric
        The label for the model.

    n_states: int
        The number of states for the model.

    n_components: int
        The number of mixture components used in the emission distribution for each state.

    covariance_type: str
        The covariance matrix type for emission distributions.

    n_seqs: int
        The number of observation sequences use to train the model.

    initial: numpy.ndarray (float)
        The initial state distribution of the model.

    transitions: numpy.ndarray (float)
        The transition matrix of the model.
    """

    def __init__(self, label, n_states, n_components=1, covariance_type='full', topology='left-right', random_state=None):
        self._val = _Validator()
        self._label = self._val.string_or_numeric(label, 'model label')
        self._label = label
        self._n_states = self._val.restricted_integer(
            n_states, lambda x: x > 0, desc='number of states', expected='greater than zero')
        self._n_components = self._val.restricted_integer(
            n_components, lambda x: x > 0, desc='number of mixture components', expected='greater than zero')
        self._covariance_type = self._val.one_of(covariance_type, ['spherical', 'diag', 'full', 'tied'], desc='covariance matrix type')
        self._val.one_of(topology, ['ergodic', 'left-right', 'linear'], desc='topology')
        self._random_state = self._val.random_state(random_state)
        self._topology = {
            'ergodic': _ErgodicTopology,
            'left-right': _LeftRightTopology,
            'linear': _LinearTopology
        }[topology](self._n_states, self._random_state)

    def set_uniform_initial(self):
        """Sets a uniform initial state distribution :math:`\\boldsymbol{\\pi}=(\\pi_1,\\pi_2,\\ldots,\\pi_M)` where :math:`\\pi_i=1/M\\quad\\forall i`."""
        self._initial = self._topology.uniform_initial()

    def set_random_initial(self):
        """Sets a random initial state distribution by sampling :math:`\\boldsymbol{\\pi}\\sim\\mathrm{Dir}(\\mathbf{1}_M)` where

        - :math:`\\boldsymbol{\\pi}=(\\pi_1,\\pi_2,\\ldots,\\pi_M)` are the initial state probabilities for each state,
        - :math:`\\mathbf{1}_M` is a vector of :math:`M` ones which are used as the concentration parameters for the Dirichlet distribution.
        """
        self._initial = self._topology.random_initial()

    def set_uniform_transitions(self):
        """Sets a uniform transition matrix according to the topology, so that given the HMM is in state :math:`i`,
        all permissible transitions (i.e. such that :math:`p_{ij}\\neq0`) :math:`\\forall j` are equally probable."""
        self._transitions = self._topology.uniform_transitions()

    def set_random_transitions(self):
        """Sets a random transition matrix according to the topology, so that given the HMM is in state :math:`i`,
        all out-going transition probabilities :math:`\\mathbf{p}_i=(p_{i1},p_{i2},\\ldots,p_{iM})` from state :math:`i`
        are generated by sampling :math:`\\mathbf{p}_i\\sim\\mathrm{Dir}(\\mathbf{1}_M)` and redistributed so that only the
        transitions permitted by the topology are non-zero."""
        self._transitions = self._topology.random_transitions()

    def fit(self, X):
        """Fits the HMM to observation sequences assumed to be labeled as the class that the model represents.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            Collection of multivariate observation sequences, each of shape :math:`(T \\times D)` where
            :math:`T` may vary per observation sequence.
        """
        X = self._val.observation_sequences(X)

        try:
            (self._initial, self._transitions)
        except AttributeError as e:
            raise AttributeError('Must specify initial state distribution and transitions before the HMM can be fitted') from e

        self._n_seqs = len(X)
        self._n_features = X[0].shape[1]

        # Initialize the GMMHMM with the specified initial state distribution and transition matrix
        self._model = hmmlearn.hmm.GMMHMM(
            n_components=self._n_states,
            n_mix=self._n_components,
            covariance_type=self._covariance_type,
            random_state=self._random_state,
            init_params='mcw', # only initialize means, covariances and mixture weights
        )
        self._model.startprob_, self._model.transmat_ = self._initial, self._transitions

        # Perform the Baum-Welch algorithm to fit the model to the observations
        self._model.fit(np.vstack(X), [len(x) for x in X])

        # Update the initial state distribution and transitions to reflect the updated parameters
        self._initial, self._transitions = self._model.startprob_, self._model.transmat_

    def forward(self, x):
        """Runs the forward algorithm to calculate the (log) likelihood of the model generating an observation sequence.

        Parameters
        ----------
        x: numpy.ndarray (float)
            An individual sequence of observations of size :math:`(T \\times D)` where
            :math:`T` is the number of time frames (or observations) and
            :math:`D` is the number of features.

        Returns
        -------
        log-likelihood: float
            The log-likelihood of the model generating the observation sequence.
        """
        try:
            self._model
        except AttributeError as e:
            raise AttributeError('The model must be fitted before running the forward algorithm') from e

        x = self._val.observation_sequences(x, allow_single=True)
        if not x.shape[1] == self._n_features:
            raise ValueError('Number of observation features must match the dimensionality of the original data used to fit the model')

        return self._model.score(x, lengths=None)

    @property
    def label(self):
        return self._label

    @property
    def n_states(self):
        return self._n_states

    @property
    def n_components(self):
        return self._n_components

    @property
    def covariance_type(self):
        return self._covariance_type

    @property
    def n_seqs(self):
        try:
            return self._n_seqs
        except AttributeError as e:
            raise AttributeError('The model has not been fitted and has not seen any observation sequences') from e

    @property
    def model(self):
        try:
            return self._model
        except AttributeError as e:
            raise AttributeError('The model must be fitted first') from e

    @property
    def initial(self):
        try:
            return self._initial
        except AttributeError as e:
            raise AttributeError('No initial state distribution has been defined') from e

    @initial.setter
    def initial(self, probabilities):
        self._topology.validate_initial(probabilities)
        self._initial = probabilities

    @property
    def transitions(self):
        try:
            return self._transitions
        except AttributeError as e:
            raise AttributeError('No transition matrix has been defined') from e

    @transitions.setter
    def transitions(self, probabilities):
        self._topology.validate_transitions(probabilities)
        self._transitions = probabilities

    def __repr__(self):
        module = self.__class__.__module__
        out = '{}{}('.format('' if module == '__main__' else '{}.'.format(module), self.__class__.__name__)
        attrs = [
            ('label', repr(self._label)),
            ('n_states', repr(self._n_states)),
            ('n_components', repr(self._n_components)),
            ('covariance_type', repr(self._covariance_type))
        ]
        try:
            self._initial
            attrs.append(('initial', 'array([...])'))
            self._transitions
            attrs.append(('transitions', 'array([...])'))
            self._model
            attrs.append(('n_seqs', repr(self._n_seqs)))
        except AttributeError:
            pass
        return out + ', '.join('{}={}'.format(name, val) for name, val in attrs) + ')'


import sys, numbers, numpy as np
from copy import copy
if sys.version_info >= (3, 3):
    from collections.abc import Iterable
else:
    from collections import Iterable

class _Validator:
    """Performs internal validations on various input types."""

    def observation_sequences(self, X, allow_single=False):
        """Validates observation sequence(s).

        Parameters
        ----------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            An individual observation sequence or a list of multiple observation sequences.

        allow_single: bool
            Whether to allow an individual observation sequence.

        Returns
        -------
        X: numpy.ndarray (float) or list of numpy.ndarray (float)
            The original input observation sequence(s) if valid.
        """
        X = copy(X)
        if isinstance(X, (list, np.ndarray) if allow_single else list):
            if isinstance(X, list):
                for i, x in enumerate(X):
                    if not isinstance(x, np.ndarray):
                        raise TypeError('Each observation sequence must be a numpy.ndarray')
                    if not x.ndim <= 2:
                        raise ValueError('Each observation sequence must be at most two-dimensional')
                    x = X[i] = (x if x.ndim == 2 else np.atleast_2d(x).T).astype(float)
                    if not x.shape[1] == X[0].shape[1]:
                        raise ValueError('Each observation sequence must have the same dimensionality')
            elif isinstance(X, np.ndarray):
                if not X.ndim <= 2:
                    raise ValueError('Observation sequence must be at most two-dimensional')
                X = (X if X.ndim == 2 else np.atleast_2d(X).T).astype(float)
        else:
            if allow_single:
                raise TypeError('Expected an individual observation sequence or a list of multiple observation sequences, each of type numpy.ndarray')
            else:
                raise TypeError('Expected a list of observation sequences, each of type numpy.ndarray')
        return X

    def observation_sequences_and_labels(self, X, y):
        """Validates observation sequences and corresponding labels.

        Parameters
        ----------
        X: list of numpy.ndarray (float)
            A list of multiple observation sequences.

        y: array-like of str/numeric
            A list of labels for the observation sequences.

        Returns
        -------
        X: list of numpy.ndarray (float)
            The original input observation sequences if valid.

        y: array-like of str/numeric
            The original input labels if valid.
        """
        self.observation_sequences(X, allow_single=False)
        self.iterable(y, 'labels')
        self.string_or_numeric(y[0], 'each label')
        if not all(isinstance(label, type(y[0])) for label in y[1:]):
            raise TypeError('Expected all labels to be of the same type')
        if not len(X) == len(y):
            raise ValueError('Expected the same number of observation sequences and labels')
        return X, y

    def integer(self, item, desc):
        """Validates an integer.

        Parameters
        ----------
        item: int
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: int
            The original input item if valid.
        """
        if not isinstance(item, int):
            raise TypeError("Expected {} to be an integer".format(desc))
        return item

    def string(self, item, desc):
        """Validates a string.

        Parameters
        ----------
        item: str
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: str
            The original input item if valid.
        """
        if not isinstance(item, str):
            raise TypeError("Expected {} to be a string".format(desc))
        return item

    def string_or_numeric(self, item, desc):
        """Validates a string or numeric type.

        Parameters
        ----------
        item: str or numeric
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: str or numeric
            The original input item if valid.
        """
        if not isinstance(item, (str, numbers.Number)):
            raise TypeError("Expected {} to be a string or numeric".format(desc))
        return item

    def boolean(self, item, desc):
        """Validates a boolean.

        Parameters
        ----------
        item: bool
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: bool
            The original input item if valid.
        """
        if not isinstance(item, bool):
            raise TypeError("Expected {} to be a boolean".format(desc))
        return item

    def one_of(self, item, items, desc):
        """Validates that an item is one of some permitted values.

        Parameters
        ----------
        item: Any
            The item to validate.

        items: array-like of Any
            The list of permitted values to check against.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: Any
            The original input item if valid.
        """
        if not item in items:
            raise ValueError('Expected {} to be one of {}'.format(desc, items))
        return item

    def restricted_integer(self, item, condition, desc, expected):
        """Validates an integer and checks that it satisfies some condition.

        Parameters
        ----------
        item: int
            The item to validate.

        condition: lambda
            A condition to check the item against.

        desc: str
            A description of the item being validated.

        expected: str
            A description of the condition, or expected value.

        Returns
        -------
        item: int
            The original input item if valid.
        """
        if isinstance(item, int):
            if not condition(item):
                raise ValueError('Expected {} to be {}'.format(desc, expected))
        else:
            raise TypeError("Expected {} to be an integer".format(desc))
        return item

    def restricted_float(self, item, condition, desc, expected):
        """Validates a float and checks that it satisfies some condition.

        Parameters
        ----------
        item: float
            The item to validate.

        condition: lambda
            A condition to check the item against.

        desc: str
            A description of the item being validated.

        expected: str
            A description of the condition, or expected value.

        Returns
        -------
        item: float
            The original input item if valid.
        """
        if isinstance(item, float):
            if not condition(item):
                raise ValueError('Expected {} to be {}'.format(desc, expected))
        else:
            raise TypeError("Expected {} to be a float".format(desc))
        return item

    def random_state(self, state):
        """Validates a random state object or seed.

        Parameters
        ----------
        state: None, int, numpy.random.RandomState
            A random state object or seed.

        Returns
        -------
        state: numpy.random.RandomState
            A random state object.
        """
        if state is None:
            return np.random.RandomState(seed=None)
        elif isinstance(state, int):
            return np.random.RandomState(seed=state)
        elif isinstance(state, np.random.RandomState):
            return state
        else:
            raise TypeError('Expected random state to be of type: None, int, or numpy.random.RandomState')

    def func(self, item, desc):
        """Validates a callable.

        Parameters
        ----------
        item: callable
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: callable
            The original input item if valid.
        """
        if callable(item):
            return item
        else:
            raise TypeError('Expected {} to be callable'.format(desc))

    def iterable(self, item, desc):
        """Validates an iterable.

        Parameters
        ----------
        item: iterable
            The item to validate.

        desc: str
            A description of the item being validated.

        Returns
        -------
        item: iterable
            The original input item if valid.
        """
        if isinstance(item, Iterable) and hasattr(item, '__len__'):
            return item
        else:
            raise TypeError("Expected {} to be an iterable".format(desc))


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import print_function
from setuptools import setup, find_packages

python_requires = '>=3.6,<3.10'
setup_requires = [
    'Cython>=0.28.5',
    'numpy>=1.17,<2',
    'scipy>=1.3,<2'
]
install_requires = [
    'numpy>=1.17,<2',
    'hmmlearn==0.2.4',
    'dtaidistance[numpy]>=2.2,<2.3',
    'scipy>=1.3,<2',
    'scikit-learn>=0.22,<1',
    'tqdm>=4.36,<5',
    'joblib>=0.14,<1'
]

VERSION = '0.10.2'

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name = 'sequentia',
    version = VERSION,
    author = 'Edwin Onuonga',
    author_email = 'ed@eonu.net',
    description = 'A machine learning interface for isolated sequence classification algorithms in Python.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/eonu/sequentia',
    project_urls = {
        'Documentation': 'https://sequentia.readthedocs.io/en/latest',
        'Bug Tracker': 'https://github.com/eonu/sequentia/issues',
        'Source Code': 'https://github.com/eonu/sequentia'
    },
    license = 'MIT',
    package_dir = {'': 'lib'},
    packages = find_packages(where='lib'),
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Natural Language :: English'
    ],
    python_requires = python_requires,
    setup_requires = setup_requires,
    install_requires = install_requires
)


import numpy as np

class _Topology:
    """Represents a topology for a HMM, imposing restrictions on the transition matrix and initial state distribution.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    def __init__(self, n_states: int, random_state: np.random.RandomState):
        self._n_states = n_states
        self._random_state = random_state

    def uniform_initial(self) -> np.ndarray:
        """Sets the initial state distribution as a discrete uniform distribution.

        Returns
        -------
        initial: :class:`numpy:numpy.ndarray` (float)
            The initial state distribution of shape `(n_states,)`.
        """
        return np.ones(self._n_states) / self._n_states

    def random_initial(self) -> np.ndarray:
        """Sets the initial state distribution by randomly sampling probabilities generated by a Dirichlet distribution.

        Returns
        -------
        initial: :class:`numpy:numpy.ndarray` (float)
            The initial state distribution of shape `(n_states,)`.
        """
        return self._random_state.dirichlet(np.ones(self._n_states), size=1).flatten()

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
        to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        raise NotImplementedError

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution - according to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The random transition matrix of shape `(n_states, n_states)`.
        """
        raise NotImplementedError

    def validate_initial(self, initial: np.ndarray) -> None:
        """Validates an initial state distribution according to the topology's restrictions.

        Parameters
        ----------
        initial: numpy.ndarray (float)
            The initial state distribution to validate.
        """
        if not isinstance(initial, np.ndarray):
            raise TypeError('Initial state distribution must be a numpy.ndarray')
        if not initial.shape == (self._n_states,):
            raise ValueError('Initial state distribution must be of shape (n_states,)')
        if not np.isclose(initial.sum(), 1):
            raise ValueError('Initial state distribution must sum to one')

    def validate_transitions(self, transitions: np.ndarray) -> None:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray (float)
            The transition matrix to validate.
        """
        if not isinstance(transitions, np.ndarray):
            raise TypeError('Transition matrix must be a numpy.ndarray')
        if not transitions.shape == (self._n_states, self._n_states):
            raise ValueError('Transition matrix must be of shape (n_states, n_states)')
        if not np.allclose(transitions.sum(axis=1), np.ones(self._n_states)):
            raise ValueError('Transition probabilities out of each state must sum to one')